"""
Provides utility functions used in the abstinence calculation
"""

from pathlib import Path
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from abstcal.calculator_error import InputArgumentError, FileExtensionError

# Use the following session to track data
# Reference: https://gist.github.com/tvst/036da038ab3e999a64497f42de966a92
import streamlit as st
from streamlit.report_thread import get_report_ctx


def read_data_from_path(filepath):
    """
    Read data from the specified path

    :param filepath: Union[str, Path, DataFrame], the path to the file to be read
        Supported file types: comma-separated, tab-separated, and Excel spreadsheet

    :return: a DataFrame
    """
    if isinstance(filepath, pd.DataFrame):
        return filepath

    path = Path(filepath)
    file_extension = path.suffix.lower()
    if file_extension == ".csv":
        df = pd.read_csv(filepath, infer_datetime_format=True)
    elif file_extension in (".xls", ".xlsx", ".xlsm", ".xlsb"):
        df = pd.read_excel(filepath, infer_datetime_format=True)
    elif file_extension == ".txt":
        df = pd.read_csv(filepath, sep='\t', infer_datetime_format=True)
    else:
        raise FileExtensionError(filepath)
    return df


def from_wide_to_long(data_filepath, data_source_type, subject_col_name="id"):
    """
    Convert data from the wide format to the long format

    :param data_filepath: the filepath to the data

    :param data_source_type: the source type of the data: visit or TLFB

    :param subject_col_name: the column name of the subject id

    :return: DataFrame, the transformed data in the long format
    """
    wide_df = read_data_from_path(data_filepath)
    if data_source_type == "tlfb":
        var_col_name, value_col_name = "date", "amount"
    else:
        var_col_name, value_col_name = "visit", "date"
    long_df = wide_df.melt(id_vars=subject_col_name, var_name=var_col_name, value_name=value_col_name)
    long_df.rename(columns={subject_col_name: "id"}, inplace=True)
    return long_df


def mask_dates(tlfb_filepath, bio_filepath, visit_filepath, reference):
    """
    Mask the dates in the visit and TLFB datasets using the anchor visit

    :param tlfb_filepath: the DataFrame object or filepath to the TLFB data (e.g., csv, xls, xlsx)

    :param visit_filepath: the DataFrame object or filepath to the visit data (e.g., csv, xls, xlsx)

    :param bio_filepath: the DataFrame object or filepath to the biochemical data (e.g., csv, xls, xlsx)

    :param reference: the anchor visit [it must be one member of visit list]
        or arbitrary date Union[str "mm/dd/yyyy", datetime.date] using which to mask the dates in these datasets

    :return: two DataFrames for visit and TLFB, respectively
    """
    visit_df = read_data_from_path(visit_filepath)

    tlfb_df = read_data_from_path(tlfb_filepath)

    tlfb_dfs = list()
    for filepath in filter(lambda x: x is not None, (tlfb_filepath, bio_filepath)):
        tlfb_df = read_data_from_path(filepath)
        tlfb_dfs.append(tlfb_df)

    if reference in visit_df['visit'].unique():
        anchor_visit = reference
        anchor_dates = visit_df.loc[visit_df['visit'] == anchor_visit, ["id", "date"]]. \
            rename(columns={"date": "anchor_date"})

        tlfb_dfs_anchored = list()
        for tlfb_df in tlfb_dfs:
            tlfb_df_anchored = tlfb_df.merge(anchor_dates, on="id")
            tlfb_df_anchored['date'] = (tlfb_df_anchored['date'] - tlfb_df_anchored['anchor_date']).map(
                lambda x: x.days if pd.notnull(x) else np.nan
            )
            tlfb_dfs_anchored.append(tlfb_df_anchored.drop("anchor_date", axis=1))

        visit_df_anchored = visit_df.merge(anchor_dates, on="id")
        visit_df_anchored['date'] = (visit_df_anchored['date'] - visit_df_anchored['anchor_date']).map(
            lambda x: x.days if pd.notnull(x) else np.nan
        )
        return tuple((*tlfb_dfs_anchored, visit_df_anchored.drop("anchor_date", axis=1)))
    else:
        try:
            reference_date = pd.to_datetime(reference)
        except TypeError:
            raise InputArgumentError("You're expecting to pass a date object or string as a reference.")
        else:
            for tlfb_df in tlfb_dfs:
                tlfb_df['date'] = (tlfb_df['date'] - reference_date).map(lambda x: x.days)
            visit_df['date'] = (visit_df['date'] - reference_date).map(lambda x: x.days)
        return tuple((*tlfb_dfs, visit_df))


def write_data_to_path(df, filepath, index=False):
    """
    Write data to the specified path

    :param df: DataFrame
    :param filepath: Union[str, Path, None], the path to the file to be read
        Supported file types: comma-separated, tab-separated, and Excel spreadsheet, if None, no writing will be
        performed

    :param index: bool, whether the index of the DataFrame will be written to the output file

    :return: a DataFrame
    """
    if filepath is None:
        return

    path = Path(filepath)
    file_extension = path.suffix.lower()
    if file_extension == ".csv":
        pd.DataFrame.to_csv(df, path, index=index)
    elif file_extension in (".xls", ".xlsx", ".xlsm", ".xlsb"):
        pd.DataFrame.to_excel(df, path, index=index)
    elif file_extension == ".txt":
        pd.DataFrame.to_csv(df, path, sep='\t', index=index)
    else:
        raise FileExtensionError(filepath)


show_figure = plt.show


class SessionState(object):
    def __init__(self, **kwargs):
        """A new SessionState object.

        Parameters
        ----------
        **kwargs : any
            Default values for the session state.

        Example
        -------
        >>> session_state = SessionState(user_name='', favorite_color='black')
        >>> session_state.user_name = 'Mary'
        ''
        >>> session_state.favorite_color
        'black'

        """
        for key, val in kwargs.items():
            setattr(self, key, val)


@st.cache(allow_output_mutation=True)
def get_session(id, **kwargs):
    return SessionState(**kwargs)


def get_saved_session(**kwargs):
    """Gets a SessionState object for the current session.

    Creates a new object if necessary.

    Parameters
    ----------
    **kwargs : any
        Default values you want to add to the session state, if we're creating a
        new one.

    Example
    -------
    >>> session_state = get(user_name='', favorite_color='black')
    >>> session_state.user_name
    ''
    >>> session_state.user_name = 'Mary'
    >>> session_state.favorite_color
    'black'

    Since you set user_name above, next time your script runs this will be the
    result:
    >>> session_state = get(user_name='', favorite_color='black')
    >>> session_state.user_name
    'Mary'

    """
    ctx = get_report_ctx()
    id = ctx.session_id
    return get_session(id, **kwargs)


__all__ = ["get_saved_session", "read_data_from_path", "from_wide_to_long", "mask_dates", "write_data_to_path"]



"""
CalculatorData
---------
A data model for creating subclasses to process timeline followback and visit data in the abstinence calculation
"""

from enum import Enum
import pandas as pd
from abstcal.calculator_error import InputArgumentError, FileFormatError


class DataImputationCode(Enum):
    RAW = 0
    IMPUTED = 1
    OVERRIDDEN = 2


class CalculatorData:
    data: pd.DataFrame
    duplicates: pd.DataFrame = None
    subject_ids: set
    _index_keys: list
    _value_key: str
    use_raw_date: bool

    @property
    def col_names(self):
        return [*self._index_keys, self._value_key]

    def check_duplicates(self, duplicate_kept="mean"):
        """
        Check duplicate records, which are identified using id and date for the TLFB data and using id and visit for
        the visit data

        :param duplicate_kept: Union["min", "max", "mean", False, None], specify the action with duplicate records
            "max": default option, keep the records with the maximal value
            "min": keep the records with the minimal value
            "mean": drop duplicate records and replace them with mean values
            False: remove all duplicate records
            None: no actions on duplicate records
        :return: a DataFrame of duplicate records
        """
        if duplicate_kept not in ("min", "max", "mean", False, None):
            raise InputArgumentError("Please specify how you deal with the duplicates, min, max, mean, False, or None.")

        duplicated_indices = self.data.duplicated(self._index_keys, keep=False)
        duplicates = self.data.loc[duplicated_indices, :].sort_values(by=self._index_keys, ignore_index=True)

        if duplicate_kept is None or duplicates.empty:
            return 0

        if duplicate_kept in ("min", "max"):
            self.data.sort_values(by=[*self._index_keys, self._value_key], inplace=True)
            self.data.drop_duplicates(self._index_keys, keep="first" if duplicate_kept == "min" else "last",
                                      inplace=True, ignore_index=True)
        elif duplicate_kept == "mean":
            df = self.data.drop_duplicates(self._index_keys, keep=False)
            duplicates_means = duplicates.groupby(self._index_keys)[self._value_key].mean().reset_index()
            self.data = pd.concat([df, duplicates_means], axis=0)
            self.data.sort_values(by=self._index_keys, inplace=True, ignore_index=True)
        else:
            self.data.drop_duplicates(self._index_keys, keep=False, inplace=True)

        return duplicates

    def drop_na_records(self):
        """
        Drop the records with any na values

        :return: int, the number of NA records that are dropped
        """
        missing_records = self.data.isnull().any(axis=1).sum()
        if missing_records:
            self.data.dropna(axis=0, inplace=True)
            self.data.reset_index(drop=True)
        return missing_records

    def validate_data(self, df):
        needed_cols_unordered = set(self.col_names)
        current_cols_unordered = set(df.columns)
        if needed_cols_unordered.issubset(current_cols_unordered):
            if self.use_raw_date:
                df['date'] = pd.to_datetime(df['date'], infer_datetime_format=True)
            if self._value_key == 'amount':
                df['amount'] = df['amount'].astype(float)
            return df.loc[:, self.col_names]
        else:
            raise FileFormatError(f'Required Columns of {self.__class__.__name__}: {self.col_names}')



"""
A web app for calculating abstinence data from TLFB and visit data using the abstcal package
"""

import base64
import datetime
import io
import os
import sys
import pandas as pd
import streamlit as st
# sys.path.append(os.getcwd())
# from abstcal.tlfb_data import TLFBData
# from abstcal.visit_data import VisitData
# from abstcal.abstinence_calculator import AbstinenceCalculator
# from abstcal.calculator_web_utils import get_saved_session
# from abstcal.abstcal_utils import from_wide_to_long, mask_dates
from abstcal import TLFBData, VisitData, AbstinenceCalculator, abstcal_utils

get_saved_session = abstcal_utils.get_saved_session
from_wide_to_long = abstcal_utils.from_wide_to_long
mask_dates = abstcal_utils.mask_dates


abstcal_version = '0.7.7'
update_date = datetime.date.today().strftime("%m/%d/%Y")
sidebar = st.sidebar
supported_file_types = ["csv", "xls", "xlsx"]

# Hide tracebacks
# sys.tracebacklimit = 0
session_state = get_saved_session(tlfb_data=None, visit_data=None)

# Shared options
duplicate_options_mapped = {
    "Keep the minimal only": "min",
    "Keep the maximal only": "max",
    "Keep the mean only": "mean",
    "Remove all duplicates": False
}
duplicate_options = list(duplicate_options_mapped)
duplicate_options_mapped_reversed = {value: key for key, value in duplicate_options_mapped.items()}

outlier_options_mapped = {
    "Don't examine outliers": None,
    "Remove the outliers": True,
    "Impute the outliers with the bounding values": False
}
outlier_options = list(outlier_options_mapped)
outlier_options_mapped_reversed = {value: key for key, value in outlier_options_mapped.items()}

# TLFB data-related params
tlfb_data_params = dict.fromkeys([
    "data",
    "use_raw_date",
    "cutoff",
    "subjects",
    "duplicate_mode",
    "imputation_last_record",
    "imputation_mode",
    "imputation_gap_limit",
    "outliers_mode",
    "allowed_min",
    "allowed_max"
])
tlfb_imputation_options_mapped = {
    "Don't impute missing records":               None,
    "Linear (a linear interpolation in the gap)": "linear",
    "Uniform (the same value in the gap)":        "uniform",
    "Specified Value":                            0
}
tlfb_imputation_options = list(tlfb_imputation_options_mapped)

# Visit data-related params
visit_data_params = dict.fromkeys([
    "data",
    "data_format",
    "use_raw_date",
    "expected_visits",
    "subjects",
    "duplicate_mode",
    "imputation_mode",
    "anchor_visit",
    "allowed_min",
    "allowed_max",
    "outliers_mode"
])
visit_data_formats = [
    "Long",
    "Wide"
]
visit_imputation_options_mapped = {
    "Don't impute dates":                                None,
    "The most frequent interval since the anchor visit": "freq",
    "The mean interval since the anchor visit":          "mean"
}
visit_imputation_options = list(visit_imputation_options_mapped)
visit_imputation_options_mapped_reversed = {value: key for key, value in visit_imputation_options_mapped.items()}

# Biochemical data-related params
bio_data_params = dict.fromkeys([
    "data",
    "cutoff",
    "overridden_amount",
    "duplicate_mode",
    "imputation_mode",
    "allowed_min",
    "allowed_max",
    "outliers_mode",
    "enable_interpolation",
    "half_life",
    "days_interpolation"
])

# Calculator-related params
abst_pp_params = dict()
abst_cont_params = dict()
abst_prol_params = dict()
abst_params_shared = dict()

calculation_assumptions_mapped = {
    "Intent-to-Treat (ITT)": "itt",
    "Responders-Only (RO)":  "ro"
}
calculation_assumptions = list(calculation_assumptions_mapped)
abst_options = [
    "Point-Prevalence",
    "Prolonged",
    "Continuous"
]


def _create_df_from_upload(file_upload):
    uploaded_df = None
    if file_upload is None:
        return uploaded_df
    uploaded_data = io.BytesIO(file_upload.getbuffer())
    try:
        uploaded_df = pd.read_csv(uploaded_data)
    except:
        uploaded_df = pd.read_excel(uploaded_data)
    finally:
        return uploaded_df


def _load_elements():
    _show_preprocessing_elements()
    st.title("Abstinence Calculator")
    _load_overview_elements()
    st.markdown("***")
    _load_tlfb_elements()
    st.markdown("***")
    _load_visit_elements()
    st.markdown("***")
    _load_cal_elements()
    st.markdown("***")
    _load_script_element()


def _show_preprocessing_elements():
    sidebar.header("Pre-Processing Tools")
    _create_data_conversion_tool()
    sidebar.markdown("***")
    _create_date_masking_tool()
    sidebar.markdown("***")
    _create_additional_visit_dates()


def _create_data_conversion_tool():
    sidebar.subheader("Data Conversion (Wide to Long)")
    sidebar.write(
        "The calculator uses datasets in the long format. Convert your wide-formatted data to the long format."
    )
    conversion_tool = sidebar.beta_expander("Conversion Tool Settings", False)
    wide_file = conversion_tool.file_uploader("Wide Formatted Data", type=supported_file_types)
    data_source_type = conversion_tool.radio("Data Source", ["Visit", "TLFB"])
    subject_col_name = conversion_tool.text_input("Subject ID Variable Name", "id")
    if data_source_type == "tlfb":
        conversion_tool.write("The resulted TLFB data will have three columns: id, date, and amount.")
    else:
        conversion_tool.write("The resulted Visit data will have three columns: id, visit, and date.")
    convert_button = conversion_tool.button("Convert")
    if convert_button:
        if wide_file:
            wide_df = _create_df_from_upload(wide_file)
            long_df = from_wide_to_long(wide_df, data_source_type.lower(), subject_col_name)
            conversion_tool.write(long_df)
            _pop_download_link(long_df, f"wide_{data_source_type}", "Long-Formatted Data",
                               container=conversion_tool)
        else:
            conversion_tool.error("Please upload your dataset first.")


def _create_date_masking_tool():
    sidebar.subheader("Date Masking Tool")
    sidebar.write(
        "Mask the dates in the Visit and TLFB data using an anchor visit or arbitrary date for all subjects"
    )
    masking_tool = sidebar.beta_expander("Masking Tool Settings", False)
    masking_tool.write("Visit Data (Long Format, 3 columns: id, visit, date)")
    visit_file = masking_tool.file_uploader("Upload Visit Data", type=supported_file_types)
    visit_df = _create_df_from_upload(visit_file)
    masking_tool.write("TLFB Data (Long Format, 3 columns: id, date, amount)")
    tlfb_file = masking_tool.file_uploader("Upload TLFB Data", type=supported_file_types)
    tlfb_df = _create_df_from_upload(tlfb_file)

    masking_tool.write("Optional Biochemical Data (Long Format, 3 columns: id, date, amount)")
    bio_file = masking_tool.file_uploader("Upload Bio Data", type=supported_file_types)
    bio_df = _create_df_from_upload(bio_file)

    masking_reference_options = ["Anchor Visit", "Arbitrary Date"]
    masking_reference_index = masking_reference_options.index(
        masking_tool.radio("Reference Using:", masking_reference_options)
    )
    reference = None
    if masking_reference_index == 0:
        if visit_df is not None:
            reference = masking_tool.selectbox("Select Visit", visit_df['visit'].unique())
            masking_tool.write(f"Each subject will use his/her {reference}'s date to calculate day counters.")
    else:
        reference = masking_tool.date_input("Arbitrary Date")
        masking_tool.write(f"All subjects will share {reference} as the reference to calculate day counters.")

    if masking_tool.button("Mask Date"):
        if visit_df is None or tlfb_df is None:
            masking_tool.error("Please update both files to mask dates consistently.")
        else:
            masked_tlfb_df, *masked_bio_df, masked_visit_df = \
                mask_dates(tlfb_df, bio_df, visit_df, reference)
            _pop_download_link(masked_tlfb_df, "masked_tlfb", "Masked TLFB Data",
                               container=masking_tool)
            if masked_bio_df:
                _pop_download_link(masked_tlfb_df[0], "masked_bio", "Masked Bio Data",
                                   container=masking_tool)
            _pop_download_link(masked_visit_df, "masked_visit", "Masked Visit Data",
                               container=masking_tool)


@st.cache(allow_output_mutation=True)
def _get_visit_dates():
    return pd.DataFrame(columns=["New", "Ref", "+ Days"])


def _create_additional_visit_dates():
    sidebar.subheader("Visit Dates Creation Tool")
    sidebar.write(
        "Create extra visit dates based on existing visit days"
    )
    visit_tool = sidebar.beta_expander("Creation Date Settings", True)
    visit_file = visit_tool.file_uploader("Upload Visit Data", type=supported_file_types, key="visit_tool_file")
    visit_df = _create_df_from_upload(visit_file)
    if (visit_df is not None) and (not visit_df.empty):
        dates_options = ["Using Actual Dates", "Using Day Counters"]
        using_date = dates_options.index(visit_tool.radio("Specify the date column's data", dates_options)) == 0
        new_visit_name = visit_tool.text_input("New Visit Name (Should be distinct)")
        reference_visit = visit_tool.selectbox("Reference Visit", visit_df['visit'].unique())
        days_to_add = visit_tool.number_input("Days to Add", value=0)
        dates_df = _get_visit_dates()
        if visit_tool.button("Add Row"):
            if new_visit_name:
                dates_df.loc[dates_df.shape[0]] = [new_visit_name, reference_visit, days_to_add]
            else:
                visit_tool.error("New visit name can't be empty.")
        dates_df.drop_duplicates('New', inplace=True, ignore_index=True)
        if not dates_df.empty:
            to_drop = visit_tool.multiselect("Remove Unwanted Rows", dates_df.index)
            if visit_tool.button("Drop Rows"):
                dates_df.drop(to_drop, inplace=True)
                dates_df.reset_index(drop=True, inplace=True)
            visit_tool.write("Overview of The Parameters of Extra Visit Dates")
            visit_tool.write(dates_df)
            if visit_tool.button("Create Output Data"):
                dates_list = list(dates_df.to_records(index=False))
                updated_visit_df = VisitData.add_additional_visit_dates(visit_df, dates_list, using_date)
                _pop_download_link(updated_visit_df, "updated_visit", "Updated Visit Data",
                                   container=visit_tool)


def _load_overview_elements():
    st.markdown("This web app calculates abstinence using the Timeline-Followback data in addiction research. No data "
                "will be saved or shared.")
    st.markdown("For advanced use cases and detailed API references, please refer to the package's "
                "[GitHub](https://github.com/ycui1-mda/abstcal) page for more information.")
    st.markdown(f"Current Version of abstcal: __{abstcal_version}__")
    st.markdown(f"Last Update Date: __{update_date}__")
    st.markdown("**Disclaimer**: Not following the steps or variation in your source data may result in incorrect "
                "abstinence results. Please verify your results for accuracy.")
    st.subheader("Basic Steps:")
    st.markdown("""
    1. Process the TLFB data in Section 1
        * The TLFB data file needs to be prepared accordingly as instructed below.
        * You can optionally integrate any biochemical measures for abstinence verification purposes.
    2. Process the Visit data in Section 2
        * The Visit data file needs to be prepared accordingly as instructed below.
    3. Calculate abstinence results in Section 3
        * It supports continuous, point-prevalence, and prolonged abstinence.
    """)

    st.subheader("Advanced Settings")
    st.markdown("If you want to re-do your calculation, please press the following button. If you need to update the "
                "uploaded files, please remove them manually and re-upload the new ones.")
    if st.button("Reset Data"):
        session_state.tlfb_data = None
        session_state.visit_data = None

    st.subheader("Automatic Code Generation (Optional Feature)")
    st.markdown("At the end of this page, after specifying the parameters (you don't have to upload any data for "
                "privacy and security concerns or other reasons), you can generate a Python script for off-line use.")
    spss_link = "https://www.ibm.com/support/knowledgecenter/en/SSLVMB_24.0.0/spss/" \
                "programmability_option/python_scripting_intro.html"
    st.markdown(f"""
    * __Python Users__: you can use the script natively in your code.
    * __R Users__: the reticulate package allows you to use Python 
    ([Use Python in R](https://cran.r-project.org/web/packages/reticulate/))
    * __Stata Users__: Stata 16.0+ has integrated Python compatibility 
    ([Call Python from Stata](https://www.stata.com/python/))
    * __SAS Users__: SAS 9.4M6 supports Python-based functionalities
    ([Using Python functions inside SAS programs]
    (https://blogs.sas.com/content/sgf/2019/06/04/using-python-functions-inside-sas-programs/))
    * __SPSS Users__: you can run Python script using the GUI 
    ([Run Python script in SPSS]({spss_link}))
    """)
    st.markdown("Non-Python users need to set up the Python environment with the necessary dependencies installed. If "
                "you're not sure about this, it's recommended that you use this web app directly. If you're able to "
                "integrate the Python script in your workflow using a different language, you're welcome to share "
                "your experience here to benefit the research community.")
    st.markdown("If you're interested in creating a package in other statistical languages (e.g., R), please feel "
                "free to do so, and we're happy to collaborate.")


def _generate_calculation_script():
    script_lines = list()
    script_lines.append("# This script is automatically generated by the abstcal web app based on user input.\n"
                        "# Please carefully review the script and adjust applicable parameters before you run it.\n"
                        "# Author: Yong Cui, Ph.D.; Email: ycui1@mdanderson.org\n"
                        "# Please feel free to contact me if you need any help.")
    script_lines.append("from abstcal import TLFBData, VisitData, AbstinenceCalculator")

    needed_tlfb_data_params = tlfb_data_params.copy()
    del needed_tlfb_data_params['data']
    script_lines.append("################################################\n"
                        "# Processing TLFB Data\n"
                        "################################################")
    script_lines.append(f'# The parameters for processing the TLFB data\n'
                        f'# Please update the parameters as applicable.\n'
                        f'tlfb_data_params = {needed_tlfb_data_params}')
    script_lines.append('# The path to the TLFB data on your computer\n'
                        'tlfb_source = "Please specify the file path to the TLFB data"')
    script_lines.append("""# Create the TLFBData instance\ntlfb_data = TLFBData(
    tlfb_source,
    tlfb_data_params["cutoff"],
    tlfb_data_params["subjects"],
    tlfb_data_params["use_raw_date"]
)""")
    script_lines.append("# Profile the data\n"
                        "tlfb_data.profile_data(tlfb_data_params['allowed_min'], tlfb_data_params['allowed_max'])")
    script_lines.append("# Drop any records with missing data\n"
                        "tlfb_data.drop_na_records()\n\n"
                        "# Remove any duplicates\n"
                        "tlfb_data.check_duplicates(tlfb_data_params['duplicate_mode'])\n\n"
                        "# Recode any outliers\n"
                        "tlfb_data.recode_outliers(tlfb_data_params['allowed_min'], tlfb_data_params['allowed_min'],"
                        "tlfb_data_params['outliers_mode'])")
    if tlfb_data_params["imputation_mode"] is not None:
        script_lines.append("""# Impute the TLFB data\nimputation_params = [
    tlfb_data_params["imputation_mode"],
    tlfb_data_params["imputation_last_record"],
    tlfb_data_params["imputation_gap_limit"]
]""")
        if bio_data_params["data"] is not None:
            script_lines.append("""# Use biochemical data\nbiochemical_data = TLFBData(
    bio_data_params["data"],
    bio_data_params["cutoff"]
)""")
            if bio_data_params["enable_interpolation"]:
                script_lines.append("""# Interpolate biochemical data\nbiochemical_data.interpolate_biochemical_data(
    bio_data_params["half_life"],
    bio_data_params["days_interpolation"]
)""")
            script_lines.append("# Clean up biochemical\nbiochemical_data.drop_na_records()\n"
                                "biochemical_data.check_duplicates()")
            script_lines.append("# Apply the biochemical data to TLFB data preps\n"
                                "imputation_params.extend((biochemical_data, "
                                "str(bio_data_params['overridden_amount'])))")
        script_lines.append("tlfb_data.impute_data(*imputation_params)")

    script_lines.append("################################################\n"
                        "# Processing Visit Data\n"
                        "################################################")
    needed_visit_data_params = visit_data_params.copy()
    del needed_visit_data_params['data']
    if not needed_visit_data_params['expected_visits']:
        needed_visit_data_params['expected_visits'] = 'infer'
    script_lines.append(f'# The parameters for processing the visit data, you need to update some of them\n'
                        f'visit_data_params = {needed_visit_data_params}')
    script_lines.append('# The path to the visit data on your computer\n'
                        'visit_source = "Please specify the file path to the visit data"')
    script_lines.append("""# Create the VisitData instance\nvisit_data = VisitData(
    visit_source,
    visit_data_params["data_format"],
    visit_data_params["expected_visits"],
    visit_data_params["subjects"],
    visit_data_params["use_raw_date"]
)""")
    script_lines.append("# Profile the data\n"
                        "visit_data.profile_data(visit_data_params['allowed_min'], visit_data_params['allowed_max'])")
    script_lines.append("# Remove any duplicates\n"
                        "visit_data.check_duplicates(visit_data_params['duplicate_mode'])")
    script_lines.append("# Recode any outliers\n"
                        "visit_data.recode_outliers(visit_data_params['allowed_min'], visit_data_params['allowed_min'],"
                        "visit_data_params['outliers_mode'])")
    script_lines.append("# Impute missing dates\nvisit_data.impute_data(visit_data_params['imputation_mode'], "
                        "anchor_visit=visit_data_params['anchor_visit'])")

    script_lines.append("################################################\n"
                        "# Abstinence Calculation\n"
                        "################################################\n"
                        "# Create the calculator using TLFB and visit data\n"
                        "calculator = AbstinenceCalculator(tlfb_data, visit_data)")

    script_lines.append(f"# The shared parameters for abstinence calculation\n"
                        f"abst_params_shared = {abst_params_shared}")
    script_lines.append("# Create a list to hold calculation results\n"
                        "calculation_results = list()")
    script_lines.append(f"# Point-prevalence abstinence\n"
                        f"# Please update the parameters for point-prevalence abstinence calculations\n"
                        f"abst_pp_params = {abst_pp_params}")
    script_lines.append("""abstinence_pp = calculator.abstinence_pp(
    abst_pp_params["visits"],
    abst_pp_params["days"],
    abst_pp_params["abst_var_names"],
    abst_params_shared["including_end"],
    abst_params_shared["mode"]
)
calculation_results.append(abstinence_pp)
""")

    script_lines.append(f"# Point-Prolonged abstinence\n"
                        f"# Please update the parameters for prolonged abstinence calculations\n"
                        f"abst_prol_params = {abst_prol_params}")
    script_lines.append("""abstinence_prol = calculator.abstinence_prolonged(
    abst_prol_params["quit_visit"],
    abst_prol_params["visits"],
    abst_prol_params["lapse_definitions"],
    abst_prol_params["grace_period"],
    abst_prol_params["abst_var_names"],
    abst_params_shared["including_end"],
    abst_params_shared["mode"]
)
calculation_results.append(abstinence_prol)
""")

    script_lines.append(f"# Continuous abstinence\n"
                        f"# Please update the parameters for continuous abstinence calculations\n"
                        f"abst_cont_params = {abst_cont_params}")
    script_lines.append("""abstinence_cont = calculator.abstinence_cont(
    abst_cont_params["start_visit"],
    abst_cont_params["visits"],
    abst_cont_params["abst_var_names"],
    abst_params_shared["including_end"],
    abst_params_shared["mode"]
)
calculation_results.append(abstinence_cont)
""")

    script_lines.append("# Write the results to files on your computer\n"
                        "# Set the file path to save the abstinence results\n"
                        "abst_filepath = 'set the file path'\n"
                        "# Set the file path to save the lapse results\n"
                        "lapse_filepath = 'set the file path'")
    script_lines.append("# Save the abstinence data\n"
                        "calculator.merge_abst_data([x[0] for x in calculation_results], abst_filepath)")
    script_lines.append("# Save the lapse data\n"
                        "calculator.merge_lapse_data([x[1] for x in calculation_results], lapse_filepath)")
    generated_script = '\n\n'.join(script_lines)
    return generated_script


def _load_tlfb_elements():
    st.header("Section 1. TLFB Data")
    st.markdown("""The dataset should have three columns: __*id*__, 
    __*date*__, and __*amount*__. The id column stores the subject ids, each of which should 
    uniquely identify a study subject. The date column stores the dates when daily substance 
    uses are collected. The amount column stores substance uses for each day. Supported file 
    formats include comma-separated (.csv), tab-delimited (.txt), and Excel spreadsheets (.xls, .xlsx).  \n\n
      \n\nid | date | amount 
    ------------ | ------------- | -------------
    1000 | 02/03/2019 | 10
    1000 | 02/04/2019 | 8
    1000 | 02/05/2019 | 12
      \n\n
    """)
    container = st.beta_container()
    uploaded_file = container.file_uploader(
        "Specify the TLFB data file on your computer.",
        ["csv", "txt", "xlsx"]
    )
    tlfb_subjects = list()
    if uploaded_file:
        data = io.BytesIO(uploaded_file.getbuffer())
        tlfb_data_params["data"] = df = pd.read_csv(data)
        container.write(df)
        tlfb_data = TLFBData(df)
        tlfb_subjects = sorted(tlfb_data.subject_ids)
    else:
        container.write("The TLFB data are shown here after loading.")

    with st.beta_expander("TLFB Data Processing Advanced Configurations"):
        st.write("1. The TLFB data's date column can use either actual dates or arbitrary day counters. Please specify "
                 "the date data type.")
        tlfb_data_params["use_raw_date"] = st.checkbox(
            "Raw dates are used.",
            value=True
        )
        if tlfb_data_params["use_raw_date"]:
            st.write("The TLFB dataset uses the actual dates.")
        else:
            st.write("The TLFB dataset uses arbitrary day counters.")
        st.markdown("***")

        st.write("2. Specify the cutoff value for abstinence")
        tlfb_data_params["cutoff"] = st.number_input(
            "Cutoff Level",
            step=None
        )
        st.write(f"Data records with a value higher than {tlfb_data_params['cutoff']} are considered non-abstinent.")
        st.markdown("***")

        st.write("3. Subjects used in the abstinence calculation.")
        use_all_subjects = st.checkbox(
            "Use all subjects in the TLFB data",
            value=True
        )
        if use_all_subjects:
            tlfb_data_params["subjects"] = "all"
        else:
            tlfb_data_params["subjects"] = st.multiselect(
                "Choose the subjects of the TLFB data whose abstinence will be calculated.",
                tlfb_subjects,
                default=tlfb_subjects
            )
        st.write(f"Subjects used in the calculation: {tlfb_data_params['subjects']}")
        st.markdown("***")

        st.write("4. TLFB Missing Data Imputation (missing data are gaps having no records between study dates)")
        imputation_summary = dict()
        imputation_mode_col, imputation_value_col = st.beta_columns(2)
        selected_imputation_mode = imputation_mode_col.selectbox(
            "Select your option",
            tlfb_imputation_options,
            index=1,
            key="tlfb_imputation_mode"
        )
        imputation_summary['Imputation Mode'] = selected_imputation_mode
        tlfb_imputation_mode = tlfb_imputation_options_mapped[selected_imputation_mode]
        if tlfb_imputation_mode == 0:
            tlfb_imputation_mode = imputation_value_col.number_input(
                "Specify the value to fill the missing TLFB records.",
                value=0
            )
            imputation_summary['Missing data will be imputed using the value'] = tlfb_imputation_mode

        if tlfb_imputation_mode is not None:
            enable_gap = st.checkbox("Set limit for the maximal gap for imputation")
            if enable_gap:
                tlfb_data_params["imputation_gap_limit"] = st.number_input(
                    "Maximal Gap for Imputation (days)",
                    value=30,
                    step=1
                )
                imputation_summary['Maximally Allowed Gap for Imputation'] = \
                    f'{tlfb_data_params["imputation_gap_limit"]} days'
            else:
                imputation_summary['Enable Gap Limit'] = \
                    "There is no set limit for the maximally allowed gap for imputation."
            enable_last_record = st.checkbox(
                "Interpolate Last Record For Each Subject",
                value=True
            )
            if enable_last_record:
                tlfb_data_params["imputation_last_record"] = st.text_input(
                    "Last Record Interpolation (fill forward or a numeric value)",
                    value="ffill"
                )
                if tlfb_data_params["imputation_last_record"] == "ffill":
                    imputation_summary["Last Record Action"] = "Be filled with the last record for each subject"
                else:
                    imputation_summary["Last Record Filled With Value"] = tlfb_data_params["imputation_last_record"]
            else:
                imputation_summary["Last Record Action"] = "Will not interpolate any records beyond the last records"
        tlfb_data_params["imputation_mode"] = tlfb_imputation_mode
        imputation_summary_text = '\n\n* '.join([f'{key}: {value}' for key, value in imputation_summary.items()])
        st.markdown(f"_Summary of the Imputation Parameters_\n\n* {imputation_summary_text}")
        st.markdown("***")

        st.write("5. TLFB Duplicate Records Action (duplicates are those with the same id and date)")
        selected_duplicate_mode = st.selectbox(
            "Select your option",
            duplicate_options,
            index=len(duplicate_options) - 2,
            key="tlfb_duplicate_mode"
        )
        tlfb_data_params["duplicate_mode"] = duplicate_options_mapped[selected_duplicate_mode]
        st.write(f"Duplicate Records: {selected_duplicate_mode}")
        st.markdown("***")

        st.write("6. TLFB Outliers Actions (outliers are those lower than the min or higher than the max)")
        selected_outlier_mode = st.selectbox(
            "Select your option",
            outlier_options,
            key="tlfb_outliers_mode"
        )
        tlfb_data_params["outliers_mode"] = outlier_options_mapped[selected_outlier_mode]
        if tlfb_data_params["outliers_mode"] is not None:
            left_col, right_col = st.beta_columns(2)
            tlfb_data_params["allowed_min"] = left_col.number_input(
                "Allowed Minimal Daily Value",
                step=None,
                value=0.0
            )
            tlfb_data_params["allowed_max"] = right_col.number_input(
                "Allowed Maximal Daily Value",
                step=None,
                value=100.0
            )
        _show_outlier_action_summary(tlfb_data_params, selected_outlier_mode)

        st.write("7. Biochemical Data for Abstinence Verification (Optional)")
        has_bio_data = st.checkbox("Integrate Biochemical Data For Abstinence Calculation")
        st.markdown("""
        If your study has collected biochemical verification data, such as carbon monoxide for smoking or breath alcohol 
        concentration for alcohol intervention, these biochemical data can be integrated into the TLFB data. 
        In this way, non-honest reporting can be identified (e.g., self-reported of no use, but biochemically un-verified), 
        the self-reported value will be overridden, and the updated record will be used in later abstinence calculation.

        Please note that the biochemical measures dataset should have the same data structure as you TLFB dataset. 
        In other words, it should have three columns: id, date, and amount.

        id | date | amount 
        ------------ | ------------- | -------------
        1000 | 02/03/2019 | 4
        1000 | 02/11/2019 | 6
        1000 | 03/04/2019 | 10
        ***
        """)
        if has_bio_data:
            bio_container = st.beta_container()
            uploaded_file = bio_container.file_uploader(
                "Specify the Biochemical data file on your computer.",
                ["csv", "txt", "xlsx"]
            )
            if uploaded_file:
                data = io.BytesIO(uploaded_file.getbuffer())
                bio_data_params["data"] = pd.read_csv(data)
                bio_container.write(bio_data_params["data"])
            else:
                bio_container.write("The Biochemical data are shown here after loading.")

            st.write("7.1. Specify the cutoff value for biochemically-verified abstinence")
            bio_data_params["cutoff"] = st.number_input(
                "Equal or below the specified value is considered abstinent.",
                value=0.0,
                step=None,
                key="bio_cutoff"
            )
            st.write(f'Data records with a value higher than {bio_data_params["cutoff"]} are considered non-abstinent.')

            st.write("7.2. Override False Negative TLFB Records")
            bio_data_params["overridden_amount"] = st.number_input(
                "Specify the TLFB amount to override a false negative TLFB record. "
                "(self-report TLFB records say abstinent, but biochemical data invalidate them).",
                value=tlfb_data_params["cutoff"] + 1
            )
            st.write(f'False negative TLFB records will be overridden with {bio_data_params["overridden_amount"]}.')

            st.write("7.3. Biochemical Data Interpolation")
            st.markdown("The calculator will estimate the biochemical levels based on the "
                        "current measures in the preceding days using the half-life.")
            bio_data_params["enable_interpolation"] = st.checkbox(
                "Enable Data Interpolation"
            )
            if bio_data_params["enable_interpolation"]:
                left_col, right_col = st.beta_columns(2)
                bio_data_params["half_life"] = left_col.number_input(
                    "Half Life of the Biochemical Measure in Days",
                    min_value=0.0,
                    value=0.25,
                    step=0.01
                )
                bio_data_params["days_interpolation"] = right_col.number_input(
                    "The Number of Days of Interpolation",
                    value=1,
                    step=1
                )
                if bio_data_params["half_life"] == 0:
                    st.error("The half life of the biochemical measure should be greater than zero.")

                st.write(f'Additional biochemical records (n={bio_data_params["days_interpolation"]}) will be '
                         f'interpolated based on a half life of {bio_data_params["half_life"]} day(s).')
            else:
                st.write("There will be no interpolations for biochemical data.")

    processed_data = st.button("Get/Refresh TLFB Data Summary")

    if processed_data or session_state.tlfb_data is not None:
        _process_tlfb_data()


def _process_tlfb_data():
    tlfb_df = tlfb_data_params["data"]
    if tlfb_df is None:
        raise ValueError("Please specify the TLFB data in the file uploader above.")

    tlfb_data = TLFBData(
        tlfb_df,
        tlfb_data_params["cutoff"],
        tlfb_data_params["subjects"]
    )
    session_state.tlfb_data = tlfb_data
    abst_params_shared["tlfb_data"] = tlfb_data
    _load_data_summary(tlfb_data, tlfb_data_params)

    tlfb_imputation_mode = tlfb_data_params["imputation_mode"]
    if tlfb_imputation_mode is not None:
        st.write("Imputation Summary")
        imputation_params = [
            tlfb_data_params["imputation_mode"],
            tlfb_data_params["imputation_last_record"],
            tlfb_data_params["imputation_gap_limit"]
        ]
        bio_messages = list()
        if bio_data_params["data"] is not None:
            bio_messages.append("Note: Biochemical Data are used for TLFB imputation")
            biochemical_data = TLFBData(
                bio_data_params["data"],
                bio_data_params["cutoff"]
            )
            bio_messages.append(f"Cutoff: {bio_data_params['cutoff']}")
            bio_messages.append(f"Interpolation: {bio_data_params['enable_interpolation']}")
            if bio_data_params["enable_interpolation"]:
                bio_messages.append(f'Half Life in Days: {bio_data_params["half_life"]}')
                bio_messages.append(f'Interpolated Days: {bio_data_params["days_interpolation"]}')
                bio_messages.append(f'Overridden Amount for False Negative TLFB Records: '
                                    f'{bio_data_params["overridden_amount"]}')
                biochemical_data.interpolate_biochemical_data(
                    bio_data_params["half_life"],
                    bio_data_params["days_interpolation"]
                )
            biochemical_data.drop_na_records()
            biochemical_data.check_duplicates()
            imputation_params.extend((biochemical_data, str(bio_data_params["overridden_amount"])))

        st.write(tlfb_data.impute_data(*imputation_params))
        if bio_messages:
            st.write("; ".join(bio_messages))
    else:
        st.write("Imputation Action: None")


def _load_visit_elements():
    st.header("Section 2. Visit Data")
    st.markdown("""It needs to be in one of the following two formats.  \n\n**The long format.** 
    The dataset should have three columns: __*id*__, __*visit*__, 
    and __*date*__. The id column stores the subject ids, each of which should uniquely 
    identify a study subject. The visit column stores the visits. The date column stores 
    the dates for the visits.  \n\nid | visit | date 
    ------------ | ------------- | -------------
    1000 | 0 | 02/03/2019
    1000 | 1 | 02/10/2019
    1000 | 2 | 02/17/2019  \n\n\n\n---
      \n**The wide format.** 
    The dataset should have the id column and additional columns 
    with each representing a visit.  \n\nid | v0 | v1 | v2 | v3 | v4 | v5
    ----- | ----- | ----- | ----- | ----- | ----- | ----- |
    1000 | 02/03/2019 | 02/10/2019 | 02/17/2019 | 03/09/2019 | 04/07/2019 | 05/06/2019
    1001 | 02/05/2019 | 02/13/2019 | 02/20/2019 | 03/11/2019 | 04/06/2019 | 05/09/2019""")
    container = st.beta_container()
    uploaded_file = container.file_uploader(
        "Specify the Visit data file on your computer.",
        ["csv", "txt", "xlsx"]
    )
    visit_data_params['data_format'] = st.selectbox("Specify the file format", visit_data_formats).lower()
    visits = list()
    if uploaded_file:
        data = io.BytesIO(uploaded_file.getbuffer())
        visit_data_params["data"] = df = pd.read_csv(data)
        container.write(df)
        visit_data = VisitData(df, visit_data_params['data_format'])
        visits = sorted(visit_data.visits)
        visit_data_params['expected_visits'] = visits
        visit_subjects = sorted(visit_data.subject_ids)
    else:
        container.write("The Visit data are shown here after loading.")

    with st.beta_expander("Visit Data Processing Advanced Configurations"):
        st.write("1. The TLFB data's date column can use either actual dates or arbitrary day counters. "
                 "Please specify the date data type.")
        visit_data_params['use_raw_date'] = st.checkbox(
            "Raw dates are used",
            True,
            key="visit_data_date_col"
        )
        if visit_data_params["use_raw_date"]:
            st.write("The visit dataset uses the actual dates.")
        else:
            st.write("The visit dataset uses arbitrary day counters.")
        st.markdown("***")

        st.write("2. Specify the expected order of the visits (for data normality check)")
        visit_data_params['expected_visits'] = st.multiselect(
            "Please adjust the order accordingly",
            visits,
            default=visits
        )
        if visit_data_params["expected_visits"]:
            st.write(f'The expected order for visits: {visit_data_params["expected_visits"]}')
        else:
            st.write("The expected order for visits will be inferred from the numeric/alphabetic values.")
        st.markdown("***")

        st.write("3. Subjects used in the abstinence calculation.")
        use_all_subjects = st.checkbox(
            "Use all subjects in the Visit data",
            value=True
        )
        if use_all_subjects:
            visit_data_params["subjects"] = "all"
        else:
            visit_data_params["subjects"] = st.multiselect(
                "Choose the subjects of the TLFB data whose abstinence will be calculated.",
                visit_subjects,
                default=visit_subjects
            )
        st.write(f"Subjects used in the calculation: {visit_data_params['subjects']}")
        st.markdown("***")

        st.write("4. Visit Missing Dates Imputation")
        imputation_summary = dict()
        selected_imputation_mode = st.selectbox(
            "Select your option",
            visit_imputation_options,
            index=1,
            key="visit_imputation_mode"
        )
        imputation_summary['Imputation Mode'] = selected_imputation_mode
        visit_data_params["imputation_mode"] = visit_imputation_options_mapped[selected_imputation_mode]
        if visit_data_params["imputation_mode"] is not None:
            visit_data_params["anchor_visit"] = st.selectbox(
                "Anchor Visit for Imputation",
                visit_data_params['expected_visits'],
                index=0
            )
            imputation_summary['Anchor Visit'] = visit_data_params["anchor_visit"]
        imputation_summary_text = '\n\n* '.join([f'{key}: {value}' for key, value in imputation_summary.items()])
        st.markdown(f"_Summary of the Imputation Parameters_\n\n* {imputation_summary_text}")
        st.markdown("***")

        st.write("5. Visit Duplicate Records Action")
        selected_duplicate_mode = st.selectbox(
            "Select your option",
            duplicate_options,
            index=len(duplicate_options) - 2,
            key="visit_duplicate_mode"
        )
        visit_data_params["duplicate_mode"] = duplicate_options_mapped[selected_duplicate_mode]
        st.write(f"Duplicate Records: {selected_duplicate_mode}")
        st.markdown("***")

        st.write("6. Visit Outliers Action (outliers are those lower than the min or higher than the max)")
        selected_outlier_mode = st.selectbox(
            "Select your option",
            outlier_options,
            key="visit_outliers_mode"
        )
        visit_data_params["outliers_mode"] = outlier_options_mapped[selected_outlier_mode]
        if visit_data_params["outliers_mode"] is not None:
            left_col, right_col = st.beta_columns(2)
            visit_data_params["allowed_min"] = left_col.date_input(
                "Allowed Minimal Visit Date",
                value=datetime.datetime.today() - datetime.timedelta(days=365 * 10)
            )
            visit_data_params["allowed_max"] = right_col.date_input(
                "Allowed Maximal Visit Date",
                value=None
            )
        _show_outlier_action_summary(visit_data_params, selected_outlier_mode)

    processed_data = st.button("Get/Refresh Visit Data Summary")

    if processed_data or session_state.visit_data is not None:
        _process_visit_data()


def _show_outlier_action_summary(data_params, selected_outlier_mode):
    if selected_outlier_mode == outlier_options[0]:
        st.write("Any potential outliers won't be examined.")
    elif selected_outlier_mode == outlier_options[1]:
        st.write(f'Outliers will be removed if they are smaller than {visit_data_params["allowed_min"]} or '
                 f'greater than {visit_data_params["allowed_max"]}')
    else:
        st.write(f'Outliers will be set to the boundary values if they are smaller than '
                 f'{visit_data_params["allowed_min"]} or greater than {visit_data_params["allowed_max"]}')
    st.markdown("***")


def _process_visit_data():
    visit_df = visit_data_params["data"]
    if visit_df is None:
        raise ValueError("Please upload your Visit data and make sure it's loaded successfully.")

    st.write("Data Overview")
    visit_data = VisitData(
        visit_df,
        visit_data_params["data_format"],
        visit_data_params["expected_visits"],
        visit_data_params["subjects"]
    )
    session_state.visit_data = visit_data
    abst_params_shared["visit_data"] = visit_data
    _load_data_summary(visit_data, visit_data_params)
    imputation_mode = visit_data_params["imputation_mode"]
    if imputation_mode is not None:
        st.write(f'Imputation Summary (Parameters: anchor visit={visit_data_params["anchor_visit"]}, '
                 f'mode={visit_imputation_options_mapped_reversed[imputation_mode]})')
        st.write(visit_data.impute_data(visit_data_params["imputation_mode"], visit_data_params["anchor_visit"]))
    else:
        st.write("Imputation Action: None")
    st.write("Visit Attendance Summary")
    st.write(visit_data.get_retention_rates())


def _load_data_summary(data, data_params):
    st.subheader("Data Overview")
    data_all_summary, data_subject_summary, grid = \
        data.profile_data(data_params["allowed_min"], data_params["allowed_max"])
    st.write(data_all_summary)
    st.write(data_subject_summary)
    st.pyplot(grid)

    na_number = data.drop_na_records()
    st.write(f"Removed Records With N/A Values Count: {na_number}")

    duplicate_mode = data_params["duplicate_mode"]
    duplicates = data.check_duplicates(duplicate_mode)
    st.write(f"Duplicate Records Count: {duplicates}; "
             f"Duplicate Records Action: {duplicate_options_mapped_reversed[duplicate_mode]}")

    outliers_mode = data_params["outliers_mode"]
    if outliers_mode is not None:
        st.write(f"Outliers Summary for Action: {outlier_options_mapped_reversed[outliers_mode]}")
        st.write(data.recode_outliers(
            data_params["allowed_min"],
            data_params["allowed_max"],
            data_params["outliers_mode"])
        )
    else:
        st.write("Outliers Action: None")


def _load_cal_elements():
    st.header("Section 3. Calculate Abstinence")
    selected_mode = st.selectbox("Abstinence Assumption Mode", calculation_assumptions)
    abst_params_shared["mode"] = calculation_assumptions_mapped[selected_mode]
    st.write(f"Abstinence Assumption: {selected_mode}")
    abst_params_shared["including_end"] = st.checkbox(
        "Including each of the visit dates as the end of the time window examined."
    )
    if abst_params_shared["including_end"]:
        st.write("The end visit's date will be included in the calculation. For example, say EOT is 07/14/2020, if the "
                 "option is checked, the TLFB records up to 07/14/2020 will be included in the calculation.")
    else:
        st.write("The end visit's date won't be included in the calculation. For example, EOT is 07/14/2020, if the "
                 "option is checked, the TLFB up to 07/13/2020 will be included in the calculation.")
    pp_col, prol_col, cont_col = columns = st.beta_columns(3)
    abst_params_list = (abst_pp_params, abst_prol_params, abst_cont_params)
    abst_var_name_options = ("Infer automatically", "Specify custom variable names")
    for abst_option, col, abst_params in zip(abst_options, columns, abst_params_list):
        col.write(abst_option)
        abst_params["visits"] = col.multiselect(
            "1. Visits for Abstinence Calculation",
            visit_data_params['expected_visits'],
            key=abst_option
        )
        if abst_params["visits"]:
            col.write(f"The following visit's pp abstinence will be calculated: {abst_params['visits']}")

        abst_var_name_option = col.selectbox(
            "2. Abstinent Variable Names",
            options=abst_var_name_options,
            key=abst_option + "_name_option"
        )
        if abst_var_name_option != abst_var_name_options[0]:
            abst_names_text = col.text_input(
                "Custom Abstinence Variable Names (enclose each name within single quotes and names should be separated"
                "by commas). Example: 'abst_var1', 'abst_var2', 'abst_var3'",
                key=abst_option + "_name")
            try:
                abst_names_list = eval(f"[{abst_names_text}]")
            except SyntaxError:
                col.error("Please follow the instruction to give names.")
            else:
                abst_params["abst_var_names"] = abst_names_list
                col.write(f"Abstinence variable names: {abst_names_list}")
        else:
            abst_params["abst_var_names"] = "infer"
            col.write("Abstinence variable names will be inferred.")

        if col is pp_col:
            days_text = col.text_input(
                "3. Specify a list of the number of days preceding the visit dates. \n"
                "Enter your options and separate them by commas. Example: 7, 14, 21"
            )
            try:
                days_list = eval(f"[{days_text}]")
            except SyntaxError:
                col.error("Please follow the instruction to give options.")
            else:
                abst_params["days"] = days_list
                col.write(f"Calculate the following days' pp: {days_list}")
        elif col is prol_col:
            abst_params["quit_visit"] = col.selectbox(
                "3. Specify the quit visit",
                visit_data_params['expected_visits']
            )
            col.write(f'Selected Quit Visit: {abst_params["quit_visit"]}')
            lapse_text = col.text_input(
                "4. Specify lapse definitions. Enter your options and separate them "
                "by commas. When lapses are not allowed, its definition is False. For all definitions, "
                "please enclose each of them within single quotes. "
                "Example: 'False', '5 cigs', '5 cigs/14 days'. "
                "See GitHub page for more details."
            )
            try:
                definitions = eval(f"[{lapse_text}]")
            except SyntaxError:
                col.error("Please follow the instruction to give options.")
            else:
                for i, definition in enumerate(definitions):
                    if definition.lower().startswith("fal"):
                        definitions[i] = False
                abst_params["lapse_definitions"] = definitions

                col.write(f"Lapse Definitions (Adjusted for syntax): {definitions}")
            abst_params["grace_period"] = col.slider(
                "5. Specify the grace period in days (default: 14 days)",
                value=14,
                min_value=1,
                max_value=100,
                step=1
            )
            col.write(f"Grace Period (days): {abst_params['grace_period']}")
        else:
            abst_params["start_visit"] = col.selectbox(
                "3. Specify the start visit",
                visit_data_params['expected_visits']
            )
            col.write(f'Start Visit of the Continuous Window: {abst_params["start_visit"]}')

    if st.button("Get Abstinence Results"):
        _calculate_abstinence()


def _calculate_abstinence():
    st.header("Calculation Results Summary")
    if session_state.tlfb_data is None or session_state.visit_data is None:
        raise ValueError("Please process the TLFB and Visit data first.")

    calculator = AbstinenceCalculator(session_state.tlfb_data, session_state.visit_data)
    calculation_results = list()
    if abst_pp_params["visits"]:
        calculation_results.append(calculator.abstinence_pp(
            abst_pp_params["visits"],
            abst_pp_params["days"],
            abst_pp_params["abst_var_names"],
            abst_params_shared["including_end"],
            abst_params_shared["mode"]
        ))
    if abst_prol_params["visits"]:
        calculation_results.append(calculator.abstinence_prolonged(
            abst_prol_params["quit_visit"],
            abst_prol_params["visits"],
            abst_prol_params["lapse_definitions"],
            abst_prol_params["grace_period"],
            abst_prol_params["abst_var_names"],
            abst_params_shared["including_end"],
            abst_params_shared["mode"]
        ))
    if abst_cont_params["visits"]:
        calculation_results.append(calculator.abstinence_cont(
            abst_cont_params["start_visit"],
            abst_cont_params["visits"],
            abst_cont_params["abst_var_names"],
            abst_params_shared["including_end"],
            abst_params_shared["mode"]
        ))
    st.subheader("Abstinence Calculated for Subjects")
    st.markdown(f"{sorted(calculator.subject_ids)}")
    abst_df = calculator.merge_abst_data([x[0] for x in calculation_results])
    st.subheader("Abstinence Data")
    st.write(abst_df)
    _pop_download_link(abst_df, "abstinence_data", "Abstinence Data", True)

    lapse_df = calculator.merge_lapse_data([x[1] for x in calculation_results])
    st.subheader("Lapse Data")
    st.write(lapse_df)
    _pop_download_link(lapse_df, "lapse_data", "Lapse Data", False)


def _load_script_element():
    st.subheader("Python Script Generation")
    st.markdown("To automatize data processing, you can generate a Python script based on the input.")
    if st.button("Generate Python Script"):
        calculation_script = _generate_calculation_script()
        st.write("You can download the script here.")
        _pop_download_link(calculation_script, 'calculate_abst', 'Python Script', file_type='py')
        st.write("Alternatively, you can copy the script directly by hitting the copy icon in the right upper corner "
                 "in the box")
        st.code(calculation_script)


def _pop_download_link(df, filename, link_name, kept_index=False, file_type='csv', container=st):
    if isinstance(df, pd.DataFrame):
        data_to_download = df.to_csv(index=kept_index)
    else:
        data_to_download = df
    b64 = base64.b64encode(data_to_download.encode()).decode()
    href = f'<a href="data:file/{file_type};base64,{b64}" download="{filename}.{file_type}">Download {link_name}</a>'
    container.markdown(href, unsafe_allow_html=True)


def _max_width_(width):
    st.markdown(
        f"""<style>.reportview-container .main .block-container{{max-width: {width}px;}}</style>""",
        unsafe_allow_html=True,
    )


if __name__ == "__main__":

    _max_width_(1200)
    _load_elements()



__version__ = '0.7.7.2'

from abstcal.tlfb_data import TLFBData
from abstcal.visit_data import VisitData
# from abstcal import abstcal_utils
from abstcal.abstinence_calculator import AbstinenceCalculator



"""
AbstinenceCalculator
---------
A data model for calculating abstinence data from TLFB and visit data
"""
from datetime import timedelta
import pandas as pd
import numpy as np
from abstcal.tlfb_data import TLFBData
from abstcal.visit_data import VisitData
from abstcal.calculator_error import InputArgumentError, _show_warning
from abstcal.abstcal_utils import write_data_to_path


class AbstinenceCalculator:
    def __init__(self, tlfb_data: TLFBData, visit_data: VisitData):
        """
        Create an instance object to calculate abstinence

        :param tlfb_data: TLFBData, the TLFB data

        :param visit_data: VisitData, the visit data

        :return: None
        """
        self.tlfb_data = tlfb_data
        self.visit_data = visit_data
        self.subject_ids = sorted(tlfb_data.subject_ids & visit_data.subject_ids)

    def check_data_availability(self):
        """
        Check the data availability by crossing the TLFB dataset and Visit dataset
        :return: DataFrame
        """
        tlfb_ids = pd.Series({subject_id: 'Yes' for subject_id in self.tlfb_data.subject_ids}, name='TLFB Data')
        visit_ids = pd.Series({subject_id: 'Yes' for subject_id in self.visit_data.subject_ids}, name='Visit Data')
        crossed_data = pd.concat([tlfb_ids, visit_ids], axis=1).fillna('No')
        freq_data = crossed_data.groupby(['TLFB Data', 'Visit Data']).size().reset_index().\
            rename({0: "subject_count"}, axis=1)
        return freq_data

    def abstinence_cont(self, start_visit, end_visits, abst_var_names='infer', including_end=False, mode="itt"):
        """
        Calculates the continuous abstinence for the time window.

        :param start_visit: the visit where the time window starts, it should be one of the visits in the visit dataset

        :param end_visits: Union[list, tuple, or Any] the visits for the end(s) of the time window,
            a list of the visits or a single visit, which should all belong to the visits in the visit dataset

        :param abst_var_names: Union[str, list], the name(s) of the abstinence variable(s)
            "infer": the default option, the name(s) will be inferred,
            Note that when specified, the number of abstinence variable names should match the number of end visits

        :param including_end: bool, whether you want to include the end visit or not, default=False

        :param mode: Union["itt", "ro"], how to calculate the abstinence, "itt"=intention-to-treat (the default) or
            "ro"=responders-only

        :return: Pandas DataFrame, subject id and abstinence results
        """
        end_visits = AbstinenceCalculator._listize_args(end_visits)
        abst_names = \
            AbstinenceCalculator._infer_abst_var_names(end_visits, abst_var_names, f'{mode}_cont_v{start_visit}')
        self.visit_data._validate_visits([start_visit, *end_visits])
        if len(end_visits) != len(abst_names):
            raise InputArgumentError("The number of abstinence variable names should match the number of visits.")

        results = []
        lapses = []
        for subject_id in self.subject_ids:
            result = [subject_id]

            start_dates = self.visit_data.get_visit_dates(subject_id, start_visit, mode=mode)
            start_date = start_dates[0] if start_dates else pd.NaT
            end_dates = self.visit_data.get_visit_dates(subject_id, end_visits, int(including_end), mode=mode)

            for end_date_i, end_date in enumerate(end_dates):
                abstinent, lapse = self._score_continuous(subject_id, start_date, end_date, mode)
                result.append(abstinent)
                AbstinenceCalculator._append_lapse_as_needed(lapses, lapse, abst_names[end_date_i])

            results.append(result)

        abstinence_df = pd.DataFrame(results, columns=['id', *abst_names]).set_index("id")
        lapses_df = pd.DataFrame(lapses, columns=['id', 'date', 'amount', 'abst_name']).\
            sort_values(by=['abst_name', 'id', 'date'])

        return abstinence_df, lapses_df

    def abstinence_pp(self, end_visits, days, abst_var_names='infer', including_end=False, mode="itt"):
        """
        Calculate the point-prevalence abstinence using the end visit's date.

        :param end_visits: The reference visit(s) on which the abstinence is to be calculated

        :param days: The number of days preceding the end visit(s), it can be a single integer, or a list/tuple of days

        :param abst_var_names: The name(s) of the abstinence variable(s), by default, the name(s) will be inferred,
            if not inferred, the number of abstinence variable names should match the number of end visits

        :param including_end: Whether you want to include the anchor visit or not, default=False

        :param mode: How you want to calculate the abstinence, "itt"=intention-to-treat (the default) or
            "ro"=responders-only

        :return: Pandas DataFrame, subject id and abstinence results
        """
        days = AbstinenceCalculator._listize_args(days)
        if any(day < 1 for day in days):
            InputArgumentError("The number of days has to be a positive integer.")
        end_visits = AbstinenceCalculator._listize_args(end_visits)
        self.visit_data._validate_visits(end_visits)

        all_abst_names = list()
        for day in days:
            abst_names = AbstinenceCalculator._infer_abst_var_names(end_visits, abst_var_names, f'{mode}_pp{day}')
            all_abst_names.extend(abst_names)

        if len(end_visits) * len(days) != len(all_abst_names):
            raise InputArgumentError("The number of abstinence variable names should be equal to the multiplication of"
                                     "the number of day conditions and the number of visits.")

        results = []
        lapses = []

        for subject_id in self.subject_ids:
            result = [subject_id]
            end_dates = self.visit_data.get_visit_dates(subject_id, end_visits, int(including_end), mode=mode)

            for day_i, day in enumerate(days):
                for end_date_i, end_date in enumerate(end_dates):
                    start_date = end_date + timedelta(days=-day)
                    abstinent, lapse = self._score_continuous(subject_id, start_date, end_date, mode)
                    result.append(abstinent)
                    abst_name = all_abst_names[len(end_dates) * day_i + end_date_i]
                    AbstinenceCalculator._append_lapse_as_needed(lapses, lapse, abst_name)

            results.append(result)

        abstinence_df = pd.DataFrame(results, columns=['id', *all_abst_names]).set_index("id")
        lapses_df = pd.DataFrame(lapses, columns=['id', 'date', 'amount', 'abst_name']).\
            sort_values(by=['abst_name', 'id', 'date'])

        return abstinence_df, lapses_df

    def abstinence_prolonged(self, quit_visit, end_visits, lapse_criterion, grace_days=14, abst_var_names='infer',
                             including_end=False, mode="itt"):
        """
        Calculate the prolonged abstinence using the time window

        :param quit_visit: The visit when the subjects are scheduled to quit smoking

        :param end_visits: The end visits for which the abstinence data are to be calculated

        :param lapse_criterion: Union[False, str, list], the criterion for finding a lapse, which is only examined
            in the time window between the date when the grace period is over and the end date
            False: lapse is not allowed
            Use amounts: it must start with a numeric value, and not end with days, such as "5 cigs", "4 drinks"
            Use days: it must end with days, such as "5 days", "7 days"
            Use amounts over time: such as "5 cigs/7 days", the time intervals are rolling windows
            Use days over time: such as "5 days/7 days", the time intervals are rolling windows
            Use multiple criteria: Combinations of any of these in a tuple or list, such as ("5 cigs", "2 days",
                "2 days/7 days", False)

        :param grace_days: The number of days for the grace period following the quit date

        :param abst_var_names: The names of the abstinence variable, by default, the name will be inferred

        :param including_end: Whether you want to include the anchor visit or not, default=False

        :param mode: How you want to calculate the abstinence, "itt"=intention-to-treat (the default) or
            "ro"=responders-only

        :return: Pandas DataFrame with two columns, subject id and abstinence result
        """
        end_visits = AbstinenceCalculator._listize_args(end_visits)
        self.visit_data._validate_visits([quit_visit, *end_visits])

        all_abst_names = list()
        criteria = AbstinenceCalculator._listize_args(lapse_criterion)
        for criterion in criteria:
            if criterion:
                parsed_criterion_len = len([x for parts in criterion.split("/") for x in parts.split()])
                if parsed_criterion_len not in (2, 4):
                    raise InputArgumentError("When lapse is allowed, you have to specify the criterion for lapses in "
                                             "strings, such as '5 cigs', '5 drinks'. To see the full list of supported"
                                             "criteria, please refer to the help menu.")
            else:
                assert criterion in (False,)
            formatted_criterion = criterion.replace(" ", "_") if isinstance(criterion, str) else criterion
            abst_names = AbstinenceCalculator.\
                _infer_abst_var_names(end_visits, abst_var_names, f'{mode}_prolonged_{formatted_criterion}')
            all_abst_names.extend(abst_names)

        if len(end_visits) * len(criteria) != len(all_abst_names):
            raise InputArgumentError("The number of abstinence variable names should be equal to the multiplication of"
                                     "the number of lapse criteria and the number of visits.")

        results = []
        lapses = []

        for subject_id in self.subject_ids:
            result = [subject_id]
            start_dates = self.visit_data.get_visit_dates(subject_id, quit_visit, grace_days, mode=mode)
            start_date = start_dates[0] if start_dates else pd.NaT
            end_dates = self.visit_data.get_visit_dates(subject_id, end_visits, int(including_end), mode=mode)

            for criterion_i, criterion in enumerate(criteria):
                for end_date_i, end_date in enumerate(end_dates):
                    abstinent = 0 if mode == 'itt' else np.nan
                    lapse = None

                    if AbstinenceCalculator._validate_dates(subject_id, start_date, end_date):
                        subject_data = self.tlfb_data.get_subject_data(subject_id, start_date, end_date, mode)
                        if not criterion:
                            abstinent, lapse = self._continuous_abst(subject_id, start_date, end_date, mode)
                        else:
                            parsed_criterion = [x for parts in criterion.split("/") for x in parts.split()]
                            if len(parsed_criterion) == 2:
                                lapse_threshold = float(parsed_criterion[0])
                                lapse_tracking = 0
                                for record in subject_data.itertuples():
                                    if parsed_criterion[-1] != "days":
                                        lapse_tracking += record.amount
                                    else:
                                        lapse_tracking += (record.amount > self.tlfb_data.abst_cutoff)
                                    if lapse_tracking >= lapse_threshold:
                                        lapse = record
                                        break
                            else:
                                assert len(parsed_criterion) == 4
                                cutoff_amount, cutoff_unit, window_amount, _ = parsed_criterion
                                cutoff_amount = float(cutoff_amount)
                                window_amount = int(float(window_amount))
                                index_list = \
                                    subject_data.index[subject_data['amount'] > self.tlfb_data.abst_cutoff].tolist()
                                for j in index_list:
                                    one_window = range(j, j + window_amount)
                                    lapse_tracking = 0
                                    for elem_i, elem in enumerate(one_window, j):
                                        if cutoff_unit == "days":
                                            lapse_tracking += elem in index_list
                                        else:
                                            if elem_i in subject_data.index:
                                                lapse_tracking += subject_data.loc[elem_i, "amount"]
                                        if lapse_tracking > cutoff_amount:
                                            if elem_i in subject_data.index:
                                                lapse = subject_data.loc[elem_i, :]
                                                break
                                    if lapse is not None:
                                        break
                            if subject_data['amount'].count() == int((end_date - start_date).days):
                                abstinent = int(lapse is None)

                    result.append(abstinent)
                    abst_name = all_abst_names[len(end_dates) * criterion_i + end_date_i]
                    AbstinenceCalculator._append_lapse_as_needed(lapses, lapse, abst_name)

            results.append(result)

        abstinence_df = pd.DataFrame(results, columns=['id', *all_abst_names]).set_index("id")
        lapses_df = pd.DataFrame(lapses, columns=['id', 'date', 'amount', 'abst_name']).\
            sort_values(by=['abst_name', 'id', 'date'])

        return abstinence_df, lapses_df

    @staticmethod
    def _listize_args(raw_args):
        if not isinstance(raw_args, (tuple, list)):
            raw_args = [raw_args]
        return raw_args

    @staticmethod
    def _infer_abst_var_names(end_visits, abst_var_names, prefix=''):
        if abst_var_names == 'infer':
            abst_names = [f"{prefix}_v{end_visit}" for end_visit in end_visits]
        else:
            abst_names = AbstinenceCalculator._listize_args(abst_var_names)
        if len(end_visits) != len(abst_names):
            raise InputArgumentError("The number of abstinence variable names should match the number of visits.")
        return abst_names

    @staticmethod
    def _validate_dates(subject_id, start_date, end_date):
        validated = False
        if start_date >= end_date:
            _show_warning(f"The end date of the time window for the subject {subject_id} {end_date} isn't later \
            than the start date {start_date}. Please verify that the visit dates are correct.")
        elif pd.NaT in (start_date, end_date):
            _show_warning(f"Subject {subject_id} is missing some of the date information.")
        else:
            validated = True
        return validated

    def _score_continuous(self, subject_id, start_date, end_date, mode):
        abstinent = 0 if mode == 'itt' else np.nan
        lapse = None
        if AbstinenceCalculator._validate_dates(subject_id, start_date, end_date):
            abstinent, lapse = self._continuous_abst(subject_id, start_date, end_date, mode)
        return abstinent, lapse

    def _continuous_abst(self, subject_id, start_date, end_date, mode):
        subject_data = self.tlfb_data.get_subject_data(subject_id, start_date, end_date, mode)
        lapse_record = None
        days_to_check = int((end_date - start_date).days)
        no_missing_data = subject_data['amount'].count() == days_to_check
        if mode == "ro" and (not no_missing_data):
            return np.nan, None
        for record in subject_data.itertuples():
            if record.amount > self.tlfb_data.abst_cutoff:
                lapse_record = record
                break
        abstinent = int(no_missing_data and (lapse_record is None))
        return abstinent, lapse_record

    @staticmethod
    def _append_lapse_as_needed(lapses, lapse, abst_name):
        if lapse is not None and len(lapse):
            lapse_id, lapse_date, lapse_amount = lapse.id, lapse.date, lapse.amount
            lapses.append((lapse_id, lapse_date, lapse_amount, abst_name))

    @staticmethod
    def calculate_abstinence_rates(dfs, filepath=None):
        """
        Calculate Abstinence Rates

        :param dfs: One or more DataFrame objects, which are abstinence data generated from the calculations

        :param filepath: Union[str, Path, None], the path to the output file if set, the default is None, no output

        :return: DataFrame containing the abstinence rates
        """
        dfs = AbstinenceCalculator._listize_args(dfs)
        abst_df = pd.concat(dfs, axis=1)
        data_rows = list()
        for column in abst_df.columns:
            abst_count = int(abst_df[column].sum())
            subject_count = abst_df[column].count()
            data_row = [column, abst_count, subject_count, f"{abst_count / subject_count:.2%}"]
            data_rows.append(data_row)

        df = pd.DataFrame(data_rows, columns=['Abstinence Name', 'Abstinent Count', 'Subject Count', 'Abstinence Rate'])
        write_data_to_path(df, filepath, True)
        return df

    @staticmethod
    def merge_abst_data(dfs, filepath=None):
        """
        Merge abstinence data and write the merged DataFrame to a file

        :param dfs: Union[list, tuple], the list of abstinence data results (DataFrame), it's also OK to have just one

        :param filepath: Union[str, Path], the output file name with a proper extension

        :return: DataFrame merged
        """
        dfs = AbstinenceCalculator._listize_args(dfs)
        merged_df = pd.concat(dfs, axis=1)
        write_data_to_path(merged_df, filepath, True)
        return merged_df

    @staticmethod
    def merge_lapse_data(dfs, filepath=None):
        """
        Merge lapses data and write the merged DataFrame to a file

        :param dfs: Union[list, tuple], the list of lapse data results (DataFrame), it's also OK to have just one

        :param filepath: Union[str, path], the output file name with a proper extension

        :return: DataFrame merged
        """
        dfs = AbstinenceCalculator._listize_args(dfs)
        merged_df = pd.concat(dfs, axis=0)
        write_data_to_path(merged_df, filepath)
        return merged_df



"""
TLFBData
---------
A data model for processing timeline followback data in the abstinence calculation
"""

from collections import namedtuple
from datetime import timedelta
from abstcal.calculator_error import InputArgumentError
from abstcal.calculator_data import CalculatorData, DataImputationCode
from abstcal.abstcal_utils import read_data_from_path
import pandas as pd
import numpy as np
import seaborn as sns

TLFBRecord = namedtuple("TLFBRecord", "id date amount imputation_code")


class TLFBData(CalculatorData):
    _index_keys = ['id', 'date']
    _value_key = 'amount'

    def __init__(self, filepath, abst_cutoff=0, included_subjects="all", use_raw_date=True):
        """
        Create the instance object for the TLFB data

        :param filepath: Union[str, path, DataFrame], the file path to the TLFB data or the created DataFrame

        :param abst_cutoff: Union[float, int], the cutoff equal to or below which is abstinent

        :param included_subjects: Union[list, tuple], the list of subject ids that are included in the dataset,
            the default option "all" means that all subjects in the dataset will be used

        :param use_raw_date: Bool, whether the raw date is used in the date column (default: True)

        """
        self.use_raw_date = use_raw_date
        df = read_data_from_path(filepath)
        self.data = self.validate_data(df)
        if included_subjects and included_subjects != "all":
            self.data = self.data.loc[self.data["id"].isin(included_subjects), :].reset_index(drop=True)

        self.subject_ids = set(self.data['id'].unique())
        self.abst_cutoff = abst_cutoff

    def profile_data(self, min_amount_cutoff=None, max_amount_cutoff=None):
        """
        Profile the TLFB data

        :param min_amount_cutoff: Union[None, int, float], default None
            The minimal amount allowed for the consumption, lower than that is considered to be an outlier
            when it's set None, outlier detection won't consider the lower bound

        :param max_amount_cutoff: Union[None, int, float], default None
            The maximal amount allowed for the consumption, higher than that is considered to be an outlier
            when it's set None, outlier detection won't consider the higher bound

        :return: Tuple, summaries for the TLFB data at the sample and subject levels
        """
        tlfb_summary_series = self._get_tlfb_sample_summary(min_amount_cutoff, max_amount_cutoff)
        tlfb_subject_summary = self._get_tlfb_subject_summary(min_amount_cutoff, max_amount_cutoff)
        grid = sns.displot(self.data['amount'])
        return tlfb_summary_series, tlfb_subject_summary, grid

    def _get_tlfb_sample_summary(self, min_amount_cutoff, max_amount_cutoff):
        record_count = self.data.shape[0]
        subject_count = self.data['id'].nunique()
        record_counts = self.data.groupby('id').size()
        record_count_min, record_count_max = record_counts.min(), record_counts.max()
        tlfb_summary = {'record_count': record_count,
                        'subject_count': subject_count,
                        'records_per_subject_mean': round(record_count / subject_count, 2),
                        'records_per_subject_range': f"{record_count_min} - {record_count_max}"}
        key_names = 'min_date max_date min_amount max_amount'.split()
        col_names = 'date date amount amount'.split()
        func_names = 'min max min max'.split()
        for key_name, col_name, func_name in zip(key_names, col_names, func_names):
            func = getattr(pd.Series, func_name)
            tlfb_summary[key_name] = func(self.data[col_name])

        max_sorted_tlfb_data = \
            self.data.sort_values(by=['amount'], ascending=False).reset_index(drop=True).iloc[0:5, :].copy()
        max_sorted_tlfb_data['date'] = max_sorted_tlfb_data['date'].dt.strftime('%m/%d/%Y')
        tlfb_summary['max_amount_5_records'] = [tuple(x) for x in max_sorted_tlfb_data.values]

        min_sorted_tlfb_data = \
            self.data.dropna().sort_values(by=['amount']).reset_index(drop=True).iloc[0:5, :].copy()
        min_sorted_tlfb_data['date'] = min_sorted_tlfb_data['date'].dt.strftime('%m/%d/%Y')
        tlfb_summary['min_amount_5_records'] = [tuple(x) for x in min_sorted_tlfb_data.values]

        count_key_names = 'missing_subject_count missing_date_count missing_amount_count duplicate_count'.split()
        columns = [self.data['id'].isna(),
                   self.data['date'].isna(),
                   self.data['amount'].isna(),
                   self.data.duplicated(self._index_keys, keep=False)]
        if min_amount_cutoff is not None:
            count_key_names.append('outlier_amount_low_count')
            columns.append(pd.Series(self.data['amount'] < min_amount_cutoff))
        if max_amount_cutoff is not None:
            count_key_names.append('outlier_amount_high_count')
            columns.append(pd.Series(self.data['amount'] > max_amount_cutoff))
        for key_name, column in zip(count_key_names, columns):
            tlfb_summary[key_name] = column.sum()
        return pd.Series(tlfb_summary)

    def _get_tlfb_subject_summary(self, min_amount_cutoff, max_amount_cutoff):
        tlfb_dates_amounts = self.data.groupby('id').agg({
            "date": ["count", "min", "max"],
            "amount": ["min", "max", "mean"]
        })
        summary_dfs = [tlfb_dates_amounts]
        col_names = ['record_count', 'date_min', 'date_max', 'amount_min', 'amount_max',
                     'amount_mean', 'duplicates_count']
        summary_dfs.append(self.data.loc[self.data.duplicated(self._index_keys, keep=False), :].groupby(
            'id')['amount'].agg('count'))
        if min_amount_cutoff is not None:
            summary_dfs.append(self.data.loc[self.data['amount'] < min_amount_cutoff, :].groupby(
                'id')['amount'].agg('count'))
            col_names.append('outliers_low_count')
        if max_amount_cutoff is not None:
            summary_dfs.append(self.data.loc[self.data['amount'] > max_amount_cutoff, :].groupby(
                'id')['amount'].agg('count'))
            col_names.append('outliers_high_count')
        summary_df = pd.concat(summary_dfs, axis=1)
        summary_df.columns = col_names
        return summary_df.fillna(0)

    def interpolate_biochemical_data(self, half_life_in_days, maximum_days_to_interpolate):
        """
        Interpolate the biochemical data

        :param half_life_in_days: Union[float, int]
            The half-life of the biochemical measure to interpolate the days preceding a non-abstinent day

        :param maximum_days_to_interpolate: int, The maximum number of days to interpolate

        :return: The number of interpolated records
        """
        if half_life_in_days <= 0 or maximum_days_to_interpolate < 1 or \
                not isinstance(maximum_days_to_interpolate, int):
            InputArgumentError("The half life of the biochemical should be greater than 0, and the maximum number of "
                               "days to interpolate should be a positive integer")

        if 'imputation_code' in self.data.columns:
            self.data = self.data[self.data['imputation_code'] == 0]

        self.data.sort_values(by=self._index_keys, inplace=True, ignore_index=True)
        self.data['imputation_code'] = DataImputationCode.RAW.value

        interpolated_records = []

        for subject_id in sorted(self.subject_ids):
            subject_data = self.data[self.data['id'] == subject_id]
            interpolated_dates = set()
            for row in subject_data.itertuples():
                current_amount = row.amount
                current_date = row.date
                for day in range(1, maximum_days_to_interpolate+1):
                    interpolated_date = current_date + timedelta(days=-day)
                    if interpolated_date in interpolated_dates or current_amount <= self.abst_cutoff:
                        continue
                    interpolated_amount = pow(2, day / half_life_in_days) * current_amount
                    interpolated_record = \
                        [subject_id, interpolated_date, interpolated_amount, DataImputationCode.IMPUTED.value]
                    interpolated_dates.add(interpolated_date)
                    interpolated_records.append(interpolated_record)

        interpolated_df = pd.DataFrame(interpolated_records, columns=[*self._index_keys, 'amount', 'imputation_code'])
        self.data = pd.concat([self.data, interpolated_df]).sort_values(by=self._index_keys, ignore_index=True)
        return len(interpolated_df)

    def recode_outliers(self, floor_amount, ceil_amount, drop_outliers=True):
        """
        Recode the abnormal data of the TLFB dataset

        :param floor_amount: Union[int, float]
            Drop records when their values are lower than the floor amount if drop_outliers is True (the default),
            otherwise outliers will be replaced with the floor amount (i.e., drop_outliers=False)

        :param ceil_amount: Union[int, float]
            Drop records when their values are higher than the ceil amount if drop_outliers is True (the default),
            otherwise outliers will be replaced with the ceil amount (i.e., drop_outliers=False)

        :param drop_outliers: bool, default is True
            Drop outliers when it's True and recode outliers to bounding values when it's False

        :return: Summary of the recoding
        """
        recode_summary = dict()
        outlier_count_low = int(pd.Series(self.data['amount'] < floor_amount).sum())
        recode_summary[f'Number of outliers (< {floor_amount})'] = outlier_count_low
        self.data['amount'] = self.data['amount'].map(
            lambda x: np.nan if drop_outliers and x < floor_amount else max(x, floor_amount))
        outlier_count_high = int(pd.Series(self.data['amount'] > ceil_amount).sum())
        recode_summary[f'Number of outliers (> {ceil_amount})'] = outlier_count_high
        self.data['amount'] = self.data['amount'].map(
            lambda x: np.nan if drop_outliers and x > ceil_amount else min(x, ceil_amount))
        if drop_outliers:
            self.drop_na_records()
        return recode_summary

    # noinspection PyTypeChecker
    def impute_data(self, impute="linear", last_record_action="ffill", maximum_allowed_gap_days=None,
                    biochemical_data=None, overridden_amount="infer"):
        """
        Impute the TLFB data

        :param impute: Union["uniform", "linear", None, int, float], how the missing TLFB data are imputed
            None: no imputation
            "uniform": impute the missing TLFB data using the mean value of the amounts before and after
            the missing interval
            "linear" (the default): impute the missing TLFB data by interpolating a linear trend based on the amounts
            before and after the missing interval
            Numeric value (int or float): impute the missing TLFB data using the specified value

        :param last_record_action: Union[int, float, "ffill", None]
            to interpolate one more record from the last record (presumably the day before the last visit), this action
            is useful when you compute abstinence data involving the last visit, which may have missing data on the TLFB
            data.
            "ffill" (the default): generate one more record with the same amount of substance use as the last record
            None: no actions with the last records
            int, float: a numeric value for interpolation all subjects' last records

        :type maximum_allowed_gap_days: None or int
        :param maximum_allowed_gap_days: When it's none, there is no limit on the length of the missing gap. If it's
            set (e.g., 90 days), when the missing gap exceeds the limit, even if the TLFB records at the start and
            end of the missing block indicate no substance use, the calculator will still impute the entire window as
            using substance.

        :type biochemical_data: TLFBData
        :param biochemical_data: The biochemical data that is used to impute the self-reported data

        :param overridden_amount: Union[float, int, 'infer']
            The default is 'infer', which means that when the particular date's record exceeds the biochemical
            abstinence cutoff, while its self-reported use is below the self-reported cutoff, the self-reported use
            is interpolated as 1 unit above the self-report cutoff. For instance, in most scenarios, the self-reported
            cutoff is 0, then the self-reported use will be interpolated as 1 when the biochemical value is above the
            cutoff while the patient's self-reported use is 0.
            You can also specify what other values to be used for the interpolation when the above described condition
            is met.

        :return: Summary of the imputation
        """
        if impute is None or str(impute).lower() == "none":
            return
        if not (impute in ("uniform", "linear") or str(impute).isnumeric()):
            raise InputArgumentError("The imputation mode can only be None, 'uniform', 'linear', "
                                     "or a numeric value.")
        if 'imputation_code' in self.data.columns:
            self.data = self.data[self.data['imputation_code'] == 0]

        self.data.sort_values(by=self._index_keys, inplace=True, ignore_index=True)
        self.data['imputation_code'] = DataImputationCode.RAW.value

        to_concat = list()

        if last_record_action is not None:
            last_records = self.data.loc[self.data.groupby('id')['date'].idxmax()].copy()
            last_records['date'] = last_records['date'].map(lambda x: x + timedelta(days=1))
            if last_record_action != 'ffill':
                last_records['amount'] = float(last_record_action)
            last_records['imputation_code'] = DataImputationCode.IMPUTED.value
            to_concat.append(last_records)

        if biochemical_data is not None:
            if "imputation_code" in biochemical_data.data.columns:
                _biochemical_data = \
                    biochemical_data.data.drop(columns="imputation_code").rename(columns={"amount": "bio_amount"})
            else:
                _biochemical_data = \
                    biochemical_data.data.rename(columns={"amount": "bio_amount"})
            _merged = self.data.merge(_biochemical_data, how="left", on=self._index_keys)
            bio_amount = (self.abst_cutoff + 1) if overridden_amount == 'infer' else float(overridden_amount)

            def interpolate_tlfb(row):
                amount = row['amount']
                imputation_code = DataImputationCode.RAW.value
                if pd.isnull(row['bio_amount']):
                    return pd.Series([amount, imputation_code])
                if row['amount'] <= self.abst_cutoff and row['bio_amount'] > biochemical_data.abst_cutoff:
                    amount = bio_amount
                    imputation_code = DataImputationCode.OVERRIDDEN.value
                return pd.Series([amount, imputation_code])

            _merged[['amount', 'imputation_code']] = _merged.apply(interpolate_tlfb, axis=1)
            self.data = _merged.drop(columns="bio_amount")

        self.data['diff_days'] = self.data.groupby(['id'])['date'].diff().map(
            lambda x: (x.days if self.use_raw_date else x) if pd.notnull(x) else 1)
        missing_data = self.data[self.data['diff_days'] > 1.0]
        imputed_records = []
        for data_row in missing_data.itertuples():
            start_data = self.data.iloc[data_row.Index - 1]
            start_record = TLFBRecord(start_data.id, start_data.date, start_data.amount, start_data.imputation_code)
            end_record = TLFBRecord(data_row.id, data_row.date, data_row.amount, start_data.imputation_code)
            imputed_records.extend(
                self._impute_missing_block(start_record, end_record, impute, maximum_allowed_gap_days))
        self.data.drop(['diff_days'], axis=1, inplace=True)
        imputed_tlfb_data = pd.DataFrame(imputed_records)

        to_concat.append(self.data)
        to_concat.append(imputed_tlfb_data)

        self.data = pd.concat(to_concat).sort_values(self._index_keys, ignore_index=True)
        impute_summary = self.data.groupby(['imputation_code']).size().reset_index(). \
            rename({0: "record_count"}, axis=1)
        impute_summary['imputation_code'] = impute_summary['imputation_code'].map(
            lambda x: DataImputationCode(x).name
        )
        return impute_summary

    def _get_missing_data(self):
        self.data['diff_days'] = self.data.groupby(['id'])['date'].diff().map(
            lambda x: (x.days if self.use_raw_date else x) if pd.notnull(x) else 1)
        return self.data[self.data['diff_days'] > 1.0]

    def _impute_missing_block(self, start_record: TLFBRecord, end_record: TLFBRecord, impute, maximum_allowed_gap_days):
        subject_id, start_date, start_amount, _ = start_record
        subject_id, end_date, end_amount, _ = end_record
        imputation_code = DataImputationCode.IMPUTED.value
        day_number = (end_date - start_date).days if self.use_raw_date else (end_date - start_date)
        imputed_records = []
        if not maximum_allowed_gap_days and day_number > maximum_allowed_gap_days:
            for i in range(1, day_number):
                imputed_date = start_date + (timedelta(days=i) if self.use_raw_date else i)
                imputed_records.append(TLFBRecord(subject_id, imputed_date, self.abst_cutoff + 1, imputation_code))
            return imputed_records

        if impute == "linear":
            m = (end_amount - start_amount) / day_number
            for i in range(1, day_number):
                imputed_date = start_date + (timedelta(days=i) if self.use_raw_date else i)
                imputed_amount = m * i + start_amount
                imputed_records.append(TLFBRecord(subject_id, imputed_date, imputed_amount, imputation_code))
        elif impute == "uniform":
            imputed_amount = np.mean([start_amount, end_amount])
            for i in range(1, day_number):
                imputed_date = start_date + (timedelta(days=i) if self.use_raw_date else i)
                imputed_records.append(TLFBRecord(subject_id, imputed_date, imputed_amount, imputation_code))
        else:
            imputed_amount = float(impute)
            for i in range(1, day_number):
                imputed_date = start_date + (timedelta(days=i) if self.use_raw_date else i)
                imputed_records.append(TLFBRecord(subject_id, imputed_date, imputed_amount, imputation_code))
        return imputed_records

    def get_subject_data(self, subject_id, start_date, end_date, mode):
        df = self.data
        if mode != "itt" and 'imputation_code' in self.data.columns:
            df = self.data.loc[self.data['imputation_code'] != DataImputationCode.IMPUTED.value, :]
        subject_data = df[(df['id'] == subject_id) &
                          (start_date <= df['date']) &
                          (df['date'] < end_date)]
        return subject_data



"""
VisitData
---------
A data model for processing visit-related data in the abstinence calculation
"""
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from datetime import timedelta
from abstcal.calculator_error import InputArgumentError, _show_warning
from abstcal.calculator_data import CalculatorData, DataImputationCode
from abstcal.abstcal_utils import read_data_from_path, write_data_to_path


class VisitData(CalculatorData):
    _index_keys = ['id', 'visit']
    _value_key = 'date'

    def __init__(self, filepath, data_format="long", expected_ordered_visits="infer", included_subjects="all",
                 use_raw_date=True):
        """
        Create the instance object for the visit data

        :param filepath: Union[str, path, DataFrame], the file path to the visit data or the DataFrame

        :param data_format: Union["long", "wide"], how the visit data are organized
            long: the data should have three columns, id, visit, and date
            wide: the data have multiple columns and one row per subject, the first column has to be named 'id'
                and the remaining columns are visits with values being the dates

        :param expected_ordered_visits: Union["infer", list, None], default "infer"
            The expected order of the visits, such that when dates are out of order can be detected
            The list of visit names should match the actual visits in the dataset
            When None, no checking will be performed when you profile the visit data
            The default is infer, and it means that the order of the visits is sorted based on its numeric or
            alphabetic order

        :param included_subjects: Union[list, tuple, set], the list of subject ids that are to be
            included in the dataset, the default option "all" means that all subjects in the dataset will be used

        :param use_raw_date: Bool, whether the raw date is used in the date column (default: True)

        """
        self.use_raw_date = use_raw_date
        df_long = read_data_from_path(filepath)
        if data_format == "wide":
            df_long = df_long.melt(id_vars="id", var_name="visit", value_name="date")
        self.data = self.validate_data(df_long)
        if included_subjects != "all":
            self.data = self.data.loc[self.data["id"].isin(included_subjects), :]
        if expected_ordered_visits is not None:
            if expected_ordered_visits != "infer":
                self.expected_ordered_visits = expected_ordered_visits
                self.data = self.data.loc[self.data["visit"].isin(expected_ordered_visits), :]
            else:
                self.expected_ordered_visits = sorted(self.data['visit'].unique())
        else:
            self.expected_ordered_visits = None

        self.visits = set(self.data['visit'].unique())
        self.subject_ids = set(self.data['id'].unique())
        self.data.reset_index(drop=True, inplace=True)

    def profile_data(self, min_date_cutoff=None, max_date_cutoff=None):
        """
        Profile the visit data

        :param min_date_cutoff: Union[None, datetime, str, int], default None
            The minimal date allowed for the visit's date, lower than that is considered to be an outlier
            When it's set None, outlier detection won't consider the lower bound
            When it's str, it should be able to be casted to a datetime object, mm/dd/yyyy, such as '10/23/2020'

        :param max_date_cutoff: Union[None, datetime, str, int], default None
            The maximal amount allowed for the consumption, higher than that is considered to be an outlier
            When it's set None, outlier detection won't consider the higher bound
            When it's str, it should be able to be casted to a datetime object, such as '10/23/2020'

        :return: Tuple, summaries of the visit data at the sample and subject level
        """
        casted_min_date_cutoff = pd.to_datetime(min_date_cutoff, infer_datetime_format=True) if self.use_raw_date \
            else min_date_cutoff
        casted_max_date_cutoff = pd.to_datetime(max_date_cutoff, infer_datetime_format=True) if self.use_raw_date \
            else max_date_cutoff
        visit_summary_series, grid = self._get_visit_sample_summary(casted_min_date_cutoff, casted_max_date_cutoff)
        visit_subject_summary = self._get_visit_subject_summary(casted_min_date_cutoff, casted_max_date_cutoff)
        return visit_summary_series, visit_subject_summary, grid

    def get_out_of_order_visit_data(self):
        """
        Get the data with out of ordered visits

        :return: DataFrame or None
        """
        if self.expected_ordered_visits is None:
            return pd.DataFrame()

        subject_summary = self._get_visit_subject_summary().reset_index()
        out_of_order_ids = \
            subject_summary.loc[subject_summary['visit_dates_out_of_order'], "id"].to_list()
        return self.data.loc[self.data['id'].isin(out_of_order_ids), :].sort_values(by=['id', 'visit'])

    def _get_visit_sample_summary(self, min_date_cutoff, max_date_cutoff):
        visit_summary = {'record_count': self.data.shape[0]}
        key_names = 'subject_count visit_count distinct_visits min_date max_date'.split()
        col_names = 'id visit visit date date'.split()
        func_names = 'nunique nunique unique min max'.split()
        for key_name, col_name, func_name in zip(key_names, col_names, func_names):
            func = getattr(pd.Series, func_name)
            visit_summary[key_name] = func(self.data[col_name])

        count_key_names = 'missing_subject_count missing_visit_count missing_date_count duplicate_count'.split()
        columns = [self.data['id'].isna(),
                   self.data['visit'].isna(),
                   self.data['date'].isna(),
                   self.data.duplicated(['id', 'visit'], keep=False)]
        if min_date_cutoff is not None:
            count_key_names.append('outlier_date_low_count')
            columns.append(pd.Series(self.data['date'] < min_date_cutoff))
        if max_date_cutoff is not None:
            count_key_names.append('outlier_date_high_count')
            columns.append(pd.Series(self.data['date'] > max_date_cutoff))
        for key_name, column in zip(count_key_names, columns):
            visit_summary[key_name] = column.sum()

        min_date_indices = self.data.groupby(['id'])['date'].idxmin()
        anchor_visit = self.data.loc[min_date_indices, 'visit'].value_counts().idxmax()
        visit_summary['anchor_visit'] = anchor_visit
        anchor_dates = self.data.loc[self.data['visit'] == anchor_visit, ['id', 'date']].copy().\
            rename({'date': 'anchor_date'}, axis=1)
        visit_data_anchored = pd.merge(self.data, anchor_dates, on='id')
        visit_data_anchored['interval_to_anchor'] = (visit_data_anchored['date'] - visit_data_anchored['anchor_date']).\
            map(lambda x: x.days if self.use_raw_date else x)
        self._visit_data = visit_data_anchored.drop(['anchor_date'], axis=1)

        grid = sns.FacetGrid(self._visit_data.loc[self._visit_data['visit'] != anchor_visit, :], col="visit")
        grid.map(plt.hist, "interval_to_anchor")

        for visit in self.visits - {anchor_visit}:
            visit_dates = self._visit_data.dropna().loc[self._visit_data['visit'] == visit, :].\
                sort_values(by=['interval_to_anchor', 'id'], ignore_index=True)
            visit_summary[f"v{visit}_v{anchor_visit}_interval_days_range"] = \
                f"{visit_dates['interval_to_anchor'].min()} - {visit_dates['interval_to_anchor'].max()}"
            visit_summary[f"v{visit}_v{anchor_visit}_interval_days_min5"] = \
                [tuple(x) for x in visit_dates.loc[0:4, ['id', 'interval_to_anchor']].values]
            rows = visit_dates.shape[0]
            visit_summary[f"v{visit}_v{anchor_visit}_interval_days_max5"] = \
                [tuple(x) for x in reversed(visit_dates.loc[rows - 5:rows - 1, ['id', 'interval_to_anchor']].values)]

        visit_record_counts = self.data.groupby(['visit'])['date'].count().to_dict()
        subject_count = len(self.subject_ids)
        visit_summary.update({
            f"visit_{visit}_attendance": f"{count} ({count / subject_count:.2%})"
            for visit, count in visit_record_counts.items()
        })
        return pd.Series(visit_summary), grid

    def get_retention_rates(self, filepath=None):
        _data = self.data
        if "imputation_code" in _data.columns:
            _data = self.data[self.data['imputation_code'] == 0]

        last_attended_counts = _data.loc[_data.groupby('id')['date'].idxmax(), "visit"].value_counts().to_dict()

        sorted_visits = self.expected_ordered_visits or sorted(self.visits)

        tracking = dict()
        remaining = len(self.subject_ids)
        for i, visit in enumerate(sorted_visits):
            losing = last_attended_counts.get(visit, 0)
            tracking[visit] = remaining
            remaining -= losing

        retention_df = pd.Series(tracking).to_frame("subject_count")
        retention_df.index.name = "visit"
        retention_df.reset_index()
        subject_count = len(self.subject_ids)
        retention_df['retention_rate'] = retention_df['subject_count'].map(
            lambda x: f"{x / subject_count:.2%}"
        )
        retention_df['attrition_rate'] = retention_df['subject_count'].map(
            lambda x: f"{(subject_count - x) / subject_count:.2%}"
        )

        retention_df['attendance_rate'] = self.data['visit'].value_counts().map(
            lambda x: f"{x / subject_count:.2%}"
        )
        write_data_to_path(retention_df, filepath, True)
        return retention_df

    def _get_visit_subject_summary(self, min_date_cutoff=None, max_date_cutoff=None):
        visit_dates_amounts = self.data.groupby('id').agg({
            "date": ["count", "min", "max"]
        })
        visit_dates_amounts['date_interval'] = \
            (visit_dates_amounts[('date', 'max')] - visit_dates_amounts[('date', 'min')]).\
                map(lambda x: x.days if self.use_raw_date else x)
        summary_dfs = [visit_dates_amounts]
        col_names = ['record_count', 'date_min', 'date_max', 'date_interval', 'duplicates_count']
        summary_dfs.append(self.data.loc[self.data.duplicated(['id', 'visit'], keep=False), :].groupby(
            'id')['date'].agg('count'))
        if min_date_cutoff is not None:
            summary_dfs.append(self.data.loc[self.data['date'] < min_date_cutoff, :].groupby(
                'id')['date'].agg('count'))
            col_names.append('outliers_date_low_count')
        if max_date_cutoff is not None:
            summary_dfs.append(self.data.loc[self.data['date'] > max_date_cutoff, :].groupby(
                'id')['date'].agg('count'))
            col_names.append('outliers_date_high_count')
        visit_dates_out_of_order = self._check_visit_order()
        if visit_dates_out_of_order is not None:
            summary_dfs.append(visit_dates_out_of_order)
            col_names.append('visit_dates_out_of_order')
        summary_df = pd.concat(summary_dfs, axis=1)
        summary_df.columns = col_names
        return summary_df.fillna(0)

    def _check_visit_order(self):
        if self.expected_ordered_visits is not None:
            _temp_data = self.data.copy()
            if isinstance(self.expected_ordered_visits, list):
                _temp_data['visit'] = self.data['visit'].map(
                    {visit: i for i, visit in enumerate(self.expected_ordered_visits)})
            else:
                _show_warning("Supported options for expected_ordered_visits are list of visits, None, and infer. "
                              "The expected visit order is inferred to check if the dates are in the correct order.")
            sorted_visit_data = _temp_data.sort_values(by=self._index_keys, ignore_index=True)
            sorted_visit_data['ascending'] = sorted_visit_data.groupby(['id'])['date'].diff().map(
                lambda x: True if pd.isnull(x) or (x.days if self.use_raw_date else x) >= 0 else False
            )
            visits_out_of_order = sorted_visit_data.groupby(['id'])['ascending'].all().map(lambda x: not x)
            if visits_out_of_order.sum() > 0:
                _show_warning(f"Please note that some subjects (n={visits_out_of_order.sum()}) appear to have their \
                visit dates out of the correct order. You can find out who they are in the visit data summary by \
                subject. Please fix them if applicable.")
            return visits_out_of_order

    def recode_outliers(self, floor_date, ceil_date, drop_outliers=True):
        """
        Recode the abnormal data of the visit dataset

        :param floor_date: date (or date-like strings: "07/15/2020")
            Drop records when their values are lower than the floor date if drop_outliers is True (the default),
            otherwise outliers will be replaced with the floor date (i.e., drop_outliers=False)

        :param ceil_date: date (or date-like strings: "07/15/2020")
            Drop records when their values are higher than the ceil date if drop_outliers is True (the default),
            otherwise outliers will be replaced with the ceil date (i.e., drop_outliers=False)

        :param drop_outliers: bool, default is True
            Drop outliers when it's True and recode outliers to bounding dates when it's False

        :return: summary of the recoding
        """
        recode_summary = dict()
        casted_floor_date = pd.to_datetime(floor_date, infer_datetime_format=True) if self.use_raw_date else floor_date
        outlier_count_low = int(pd.Series(self.data['date'] < casted_floor_date).sum())
        outliers_low_key = f"Number of outliers " \
                           f"(< {casted_floor_date.strftime('%m/%d/%Y') if self.use_raw_date else casted_floor_date})"
        recode_summary[outliers_low_key] = outlier_count_low
        self.data['date'] = self.data['date'].map(
            lambda x: np.nan if drop_outliers and x < casted_floor_date else max(x, casted_floor_date))
        casted_ceil_date = pd.to_datetime(ceil_date, infer_datetime_format=True) if self.use_raw_date else ceil_date
        outlier_count_high = int(pd.Series(self.data['date'] > casted_ceil_date).sum())
        outliers_high_key = f"Number of outliers " \
                            f"(> {casted_ceil_date.strftime('%m/%d/%Y') if self.use_raw_date else casted_ceil_date})"
        recode_summary[outliers_high_key] = outlier_count_high
        self.data['date'] = self.data['date'].map(
            lambda x: np.nan if drop_outliers and x > casted_ceil_date else min(x, casted_ceil_date))
        if drop_outliers:
            self.drop_na_records()
        return recode_summary

    # noinspection PyTypeChecker
    def impute_data(self, impute='freq', anchor_visit='infer'):
        """
        Impute any missing visit data.

        :type anchor_visit: object
        :param anchor_visit: The anchor visit, which needs to exist for all subjects and the date of which will be
            used to impute the missing visit dates

        :param impute: Union["freq", "mean", None, dict], how the missing visit data are imputed
        None: no imputation
        "freq": use the most frequent, see below for further clarification
        "mean": use the average, see below for further clarification
            Both imputation methods assume that the visit structure is the same among all subjects. It will first find
            the earliest visit as the anchor date, impute any missing visit dates either using the average or the most
            frequent interval. Please note the anchor dates can't be missing.
        dict: A dictionary object mapping the number of days since the anchor visit

        :return DataFrame
        """

        if impute is None or str(impute).lower() == "none":
            return
        if impute not in ('freq', 'mean') and not isinstance(impute, dict):
            raise InputArgumentError(
                'You can only specify the imputation method to be "freq" or "mean". Alternatively, '
                'you can specify a dictionary object mapping the number of days since the anchor '
                'visit.'
            )

        if 'imputation_code' in self.data.columns:
            self.data = self.data[self.data['imputation_code'] == 0]

        self.data = self.data.sort_values(by=self._index_keys, ignore_index=True)
        if anchor_visit == 'infer' or anchor_visit is None:
            min_date_indices = self.data.groupby(['id'])['date'].idxmin()
            anchor_visit = self.data.loc[min_date_indices, 'visit'].value_counts().idxmax()

        visit_ids = sorted(self.subject_ids)
        anchor_ids = set(self.data.loc[self.data['visit'] == anchor_visit, 'id'].unique())
        missing_anchor_ids = set(visit_ids) - anchor_ids
        if missing_anchor_ids:
            message = f"Subjects {missing_anchor_ids} are missing anchor visit {anchor_visit}. " \
                      f"There might be problems calculating abstinence data for these subjects."
            _show_warning(message)
        ids_s = pd.Series(visit_ids, name='id')
        anchor_dates = self.data.loc[self.data['visit'] == anchor_visit, ['id', 'date']].\
            rename({'date': 'anchor_date'}, axis=1)
        anchor_df = pd.merge(ids_s, anchor_dates, how='outer', on='id')

        df_anchor = anchor_df.copy()
        df_anchor['visit'] = anchor_visit
        df_anchor['imputed_date'] = df_anchor['date'] = df_anchor['anchor_date']
        imputed_visit_dfs = [df_anchor]

        for visit in self.visits - {anchor_visit}:
            visit_dates = self.data.loc[self.data['visit'] == visit, ['id', 'date']]
            df_visit = pd.merge(anchor_df, visit_dates, how='outer', on='id')
            days_diff = (df_visit['date'] - df_visit['anchor_date']).map(
                lambda day_diff: (day_diff.days if self.use_raw_date else day_diff)if pd.notnull(day_diff) else np.nan)
            if impute == 'freq':
                used_days_diff = days_diff.value_counts().idxmax()
            elif impute == 'mean':
                used_days_diff = int(days_diff.mean())
            else:
                used_days_diff = impute[visit]

            def impute_date(x):
                if pd.notnull(x['date']):
                    return x['date']
                imputed_date = x['anchor_date'] + \
                               (timedelta(days=used_days_diff) if self.use_raw_date else used_days_diff)
                return imputed_date

            df_visit['imputed_date'] = df_visit.apply(impute_date, axis=1)
            df_visit['visit'] = visit
            imputed_visit_dfs.append(df_visit)

        visit_data_imputed = pd.concat(imputed_visit_dfs)
        visit_data_imputed['imputation_code'] = \
            visit_data_imputed['date'].isnull().map(int)
        self.data = visit_data_imputed.drop(['date', 'anchor_date'], axis=1). \
            rename({'imputed_date': 'date'}, axis=1).sort_values(by=self._index_keys, ignore_index=True)
        impute_summary = self.data.groupby(['imputation_code']).size().reset_index(). \
            rename({0: "record_count"}, axis=1)
        impute_summary['imputation_code'] = impute_summary['imputation_code'].map(
            lambda x: DataImputationCode(x).name
        )
        return impute_summary

    def _validate_visits(self, visits_to_check):
        if not set(visits_to_check).issubset(self.visits):
            raise InputArgumentError(
                f"Some visits are not in the visit list {self.visits}. Please check your arguments."
            )

    def get_visit_dates(self, subject_id, visit_names, increment_days=0, mode='itt'):
        """
        Retrieve the visit data for any given subject and visits

        :param subject_id: the subject id

        :param visit_names: the names of the visits

        :param increment_days: the number of day intervals to apply to the retrieved dates

        :param mode: ITT or RO, if ITT, the data will include imputed dates, otherwise, only raw data are used

        :return: List, the dates that matching the query criteria
        """

        used_data = self.data
        if mode != 'itt':
            used_data = self.data.loc[self.data['imputation_code'] == DataImputationCode.RAW.value, :]
        if not isinstance(visit_names, list):
            visit_names = [visit_names]
        visit_dates_cond = (used_data['id'] == subject_id) & (used_data['visit'].isin(visit_names))
        dates = list(used_data.loc[visit_dates_cond, 'date'])
        if increment_days:
            dates = [date + (timedelta(days=increment_days) if self.use_raw_date else increment_days) for date in dates]
        return dates

    @staticmethod
    def add_additional_visit_dates(visit_filepath, visit_dates, use_raw_date):
        """
        Add additional visit with dates using existing visits

        :param visit_filepath: Union[Path, str, DataFrame], the filepath to the visit data, or DataFrame of the visit data
            in the long format

        :param visit_dates: list[Tuple], the list of tuples, with each tuple having three items: the new visit name,
            the reference visit, the number of days to add (if you subtract the number of days, set it to negative)

        :param use_raw_date: whether the dates in the date column uses the raw dates (True), when False it means the dates
            are day counters

        :return: DataFrame
        """
        visit_df = CalculatorData.read_data_from_path(visit_filepath)
        visit_wide = visit_df.pivot(index="id", columns="visit", values="date").reset_index()
        for new_visit, reference_visit, days in visit_dates:
            visit_wide[new_visit] = visit_wide[reference_visit] + (timedelta(days=days) if use_raw_date else days)
        return visit_wide.melt(id_vars="id", var_name="visit", value_name="date") \
            .sort_values(by=['id', 'visit', 'date'], ignore_index=True)


# Aliases for different preferences
TimePointData = MilestoneData = VisitData




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="abstcal",
    # version="0.7.5",
    author="Yong Cui",
    author_email="ycui1@mdanderson.org",
    description="Calculate abstinence using the timeline followback data in substance research.",
    install_requires=['pandas', 'numpy', 'matplotlib', 'seaborn', 'streamlit', 'openpyxl', 'xlrd'],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ycui1-mda/abstcal",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)
